import numpy as np
import pandas as pd
import copy

import bluebelt.data.resolution
import bluebelt.data.projection

import bluebelt.analysis.datetime
import bluebelt.analysis.planning
import bluebelt.analysis.pattern
import bluebelt.analysis.performance

import bluebelt.statistics.std
import bluebelt.statistics.hypothesis_testing

import bluebelt.helpers.check as check

import bluebelt.helpers.holidays
from bluebelt.helpers.decorators.docstring import docstring
from bluebelt.core.filter import _filter

import bluebelt.graph.graph

import datetime


class Workload:
    def __init__(self, pandas_obj):
        self._org_obj = copy.deepcopy(pandas_obj)
        self._obj = pandas_obj
        self.log = []

        self.filter = self.filter(self._obj)
        self.pattern = self.pattern(self._obj)
        self.planning = self.planning(self._obj)
        self.forecast = self.forecast(self._obj)
        self.stats = self.stats(self._obj)
        self.performance = self.performance(self._obj)

    def __repr__(self):
        return self._obj.__repr__()

    def restore(self):
        # TO DO really needs some work
        # ugly coding

        # maybe
        # self = Workload(self._org_obj)

        attrs = list(self.__dict__.keys())
        for attr in attrs:
            if attr not in ['_org_obj', 'log', 'filter', 'pattern', 'planning', 'forecast', 'stats', 'performance']:
                delattr(self, attr)
        self._obj = self._org_obj


    # TO DO: how?!
    @_filter
    class filter:
        def __init__(self, pandas_obj):
            self._obj = pandas_obj

    def weekend(
        self, drop=None, fill_value=np.nan, inplace=None, weekend_days=[5, 6]
    ):
        """
        Fill all non-weekend data with the fill_value. If fill_value=None the
        Series or DataFrame will contain only the weekend days.

        The Series or DataFrame must have a DatetimeIndex

        Parameters
        ----------
        self : Series or DataFrame
        fill_value: object or None, default numpy.nan
        drop: boolean, default None
            if True the filtered values will be dropped
            if drop is None global settings will be applied, default True
        inplace: boolean, default None
            if True the function will be on the object itself
            if False the function will return a new object
            if inplace is None global settings will be applied, default True
        weekend_days: list of int, default [5, 6]

        Returns
        -------
        Series or DataFrame
        """
        # settings
        drop = drop or bluebelt.config("drop")
        inplace = inplace or bluebelt.config("inplace")
        
        # checks
        check.has_datetimeindex(self._obj)
        series = copy.deepcopy(self._obj) if not inplace else self._obj

        
        if fill_value is not None and not drop:
            series.loc[~series.index.weekday.isin(weekend_days)] = fill_value
        else:
            series = series[series.index.weekday.isin(weekend_days)]

        # only if Workload Class is used directly modify _obj
        if self.__class__.__name__ == "Workload":
            self._obj = series
        
        return series

    def not_weekend(
        self, drop=None, fill_value=np.nan, inplace=None, weekend_days=[5, 6]
    ):
        """
        Fill all weekend data with the fill_value. If fill_value=None the
        Series or DataFrame will contain only the non-weekend days.

        The Series or DataFrame must have a DatetimeIndex

        Parameters
        ----------
        self : Series or DataFrame
        fill_value: object or None, default numpy.nan
        drop: boolean, default None
            if drop is None global settings will be applied, default True
        inplace: boolean, default None
            if inplace is None global settings will be applied, default True
        weekend_days: list of int, default [5, 6]

        Returns
        -------
        Series or DataFrame
        """

        # settings
        drop = drop or bluebelt.config("drop")
        inplace = inplace or bluebelt.config("inplace")
        
        # checks
        check.has_datetimeindex(self._obj)
        series = copy.deepcopy(self._obj) if not inplace else self._obj

        if fill_value is not None and not drop:
            series.loc[series.index.weekday.isin(weekend_days)] = fill_value
        else:
            series = series[~series.index.weekday.isin(weekend_days)]

        if inplace:
            return
        else:
            return series

    weekdays = not_weekend

    def holiday(self, skip=None, drop=None, fill_value=np.nan, inplace=None):
        """
        Fill all non-holiday data with the fill_value. If fill_value=None the
        Series or DataFrame will contain only the holiday days.

        The Series or DataFrame must have a DatetimeIndex

        Parameters
        ----------
        self : Series or DataFrame
        fill_value: object or None, default numpy.nan
        drop: boolean, default None
            if drop is None global settings will be applied, default True
        inplace: boolean, default None
            if inplace is None global settings will be applied, default True
        
        Returns
        -------
        Series or DataFrame
        """

        # settings
        drop = drop or bluebelt.config("drop")
        inplace = inplace or bluebelt.config("inplace")
        
        # checks
        check.has_datetimeindex(self._obj)
        series = copy.deepcopy(self._obj) if not inplace else self._obj
        holidays = bluebelt.helpers.holidays.get_holidays(
            series.index.year.unique().to_list(), skip=skip
        )

        if fill_value is not None and not drop:
            series.loc[~series.index.isin(holidays)] = fill_value
        else:
            series = series[series.index.isin(holidays)]

        if inplace:
            return
        else:
            return series

    holidays = holiday

    def not_holiday(self, skip=None, drop=None, fill_value=np.nan, inplace=None):
        """
        Fill all holiday data with the fill_value. If fill_value=None the
        Series or DataFrame will contain only the non-holiday days.

        The Series or DataFrame must have a DatetimeIndex

        Parameters
        ----------
        self : Series or DataFrame
        fill_value: object or None, default numpy.nan
        drop: boolean, default None
            if drop is None global settings will be applied, default True
        inplace: boolean, default None
            if inplace is None global settings will be applied, default True
        
        Returns
        -------
        Series or DataFrame
        """

        # settings
        drop = drop or bluebelt.config("drop")
        inplace = inplace or bluebelt.config("inplace")
        
        # checks
        check.has_datetimeindex(self._obj)
        series = copy.deepcopy(self._obj) if not inplace else self._obj
        holidays = bluebelt.helpers.holidays.get_holidays(
            series.index.year.unique().to_list(), skip=skip
        )

        if fill_value is not None and not drop:
            series.loc[series.index.isin(holidays)] = fill_value
        else:
            series = series[~series.index.isin(holidays)]

        if inplace:
            return
        else:
            return series

    not_holidays = not_holiday

    def workdays(
        self, drop=None, fill_value=np.nan, inplace=None, weekend_days=[5, 6]
    ):
        """
        Fill all holiday and weekend data with the fill_value. If fill_value=None
        the Series or DataFrame will contain only the working days.

        The Series or DataFrame must have a DatetimeIndex

        Parameters
        ----------
        self : Series or DataFrame
        fill_value: object or None, default numpy.nan
        drop: boolean, default None
            if drop is None global settings will be applied, default True
        inplace: boolean, default None
            if inplace is None global settings will be applied, default True
        weekend_days: list of int, default [5, 6]

        Returns
        -------
        Series or DataFrame
        """

        # settings
        drop = drop or bluebelt.config("drop")
        inplace = inplace or bluebelt.config("inplace")
        
        # checks
        check.has_datetimeindex(self._obj)
        series = copy.deepcopy(self._obj) if not inplace else self._obj
        holidays = bluebelt.helpers.holidays.get_holidays(
            series.index.year.unique().to_list()
        )

        if fill_value is not None and not drop:
            series.loc[series.index.weekday.isin(weekend_days)] = fill_value
            series.loc[series.index.isin(holidays)] = fill_value
        else:
            series = series[~series.index.weekday.isin(weekend_days)]
            series = series[~series.index.isin(holidays)]

        if inplace:
            return
        else:
            return series

    def fill_na(self):
        """
        Fill all missing values of a pandas Series or DataFrame with the
        interpolated values of the previous and next weeks weekday.

        Parameters
        ----------
        self: Series or DataFrame

        Returns
        -------
        pandas.Series or pandas.DataFrame
        """

        if isinstance(self._obj, pd.Series):

            # transform to dataframe
            rule = datetime.timedelta(weeks=1)
            origin = bluebelt.data.resolution._get_origin(self._obj, rule)

            data = self._obj.resample(rule=rule, origin=origin)
            data = data.apply(
                lambda x: pd.DataFrame(
                    index=x.index.isocalendar().day,
                    data=x.values,
                ).transpose()
            )
            data = data.droplevel(-1)

            # interpolate
            result = pd.DataFrame(index=data.index)
            for column in data:
                series = data[column]
                mask = np.isnan(series)
                series[mask] = np.interp(
                    np.flatnonzero(mask), np.flatnonzero(~mask), series[~mask]
                )
                result[column] = series

            # finish
            result = result.stack()
            result = pd.Series(
                index=self._obj.index,
                data=result.values,
                name=self._obj.name,
            )

            return result

        else:
            # so it is a DataFrame
            result = pd.DataFrame(index=self._obj.index)

            for column in self._obj:
                result[column] = Workload.fill_na( #bluebelt.core.workload.Workload(
                    self._obj[column]
                ).fill_na()

            return result

    def resample(self, rule="w", **kwargs):
        """
        Resample time-series data.
        The object must have a DatetimeIndex

        Parameters
        ----------
        self: Series or DataFrame
        rule: DateOffset, Timedelta or str, default 'week'
            The offset string or object representing target conversion.
            e.g. '1w', 'week', '4d', '4 days'
        kwargs: other pandas.Series.resample keywords

        Returns
        -------
        pandas.core.Resampler

        Methods
        -------
        sum: Calculate the sum of all the values in each group.
        mean: Calculate the mean of all the values in each group.
        var: Calculate the variance of all the values in each group.
        std: Calculate the standard deviation of all the values in each group.
        min: Calculate the minimum value of all the values in each group.
        max: Calculate the maximum value of all the values in each group.
        count: Calculate the count of all the values in each group.
        value_range: Calculate the value range of all the values in each group.
        subsize: Calculate the count of subgroups with a size that fit in each
            group.
        diff_quantity: Calculate the difference of the sum of all values
            in adjecent groups of the resampled DataFrame.
        diff_distribution: Calculate the difference in weekday distribution
            from week to week in from a resampled DataFrame. The rule for
            resampling must be '1w' or equivalent
        diff_skills: Calculate the difference in skills distribution from week
            to week in from a resampled DataFrame. The rule for resampling must
            be '1w' or equivalent.

        """
        check.has_datetimeindex(self._obj)
        return bluebelt.data.resolution.Resample(self._obj, rule, **kwargs)

    def flatten(self, rule="w", **kwargs):
        """
        Flatten time-series data.
        The object must have a DatetimeIndex

        Parameters
        ----------
        self: Series or DataFrame
        rule: DateOffset, Timedelta or str, default 'week'
            The offset string or object representing target conversion.
            e.g. '1w', 'week', '4d', '4 days'
        kwargs: other pandas.Series.resample keywords

        Returns
        -------
        Series

        """
        check.has_datetimeindex(self._obj)
        return bluebelt.data.resolution.Flatten(self._obj, rule, **kwargs)

    @docstring(bluebelt.data.projection.project)
    def project(self, year=None, adjust_holidays=True):
        return bluebelt.data.projection.project(
            self._obj, year=year, adjust_holidays=adjust_holidays
        )

    # def peaks(self, hours=8, days=26, times=10, inplace=False):
    #     # TO DO 
    #     # is this needed? and how?
    #     # 
    #     series = copy.deepcopy(self._obj)

    #     for employee in range(times):
    #         series.loc[series.nlargest(days).index] = series.loc[series.nlargest(days).index] - hours

    #     return self._obj - series
    
    # def subtract(self, hours=8, days=26, times=10, inplace=False):
    #     # TO DO 
    #     # make pretty
    #     # find good name
    #     # option: combine with peaks
    #     if inplace:
    #         series = self._obj
    #     else:
    #         series = copy.deepcopy(self._obj)

    #     for employee in range(times):
    #         series.loc[series.nlargest(days).index] = series.loc[series.nlargest(days).index] - hours

    #     return series

    class pattern:
        def __init__(self, pandas_obj):
            self._obj = pandas_obj

        def polynomial(
            self,
            shape=(0, 6),
            validation="rsq",
            threshold=0.05,
            confidence=0.8,
            outlier_sigma=2,
            adjust=True,
            #split=False,
        ):

            """
            Find the polynomial of an object.

            Parameters
            ----------
            series: pandas.Series
            shape: int or tuple, default (0, 6)
                when an int is provided the polynomial is provided as n-th degree polynomial
                when a tuple is provided the function will find an optimised polynomial between first and second value of the tuple
            validation: string, default p_val
                validation type for shape tuple
                p_val: test for normal distribution of the residuals
                rsq: check for improvement of the rsq value
            threshold: float, default 0.05
                the threshold for normal distribution test or rsq improvement
            confidence: float, default 0.8
                the bound confidence
            outlier_sigma: float, default 2
                outliers are datapoints outside the outlier_sigma fraction
            adjust: boolean, default True
                adjust polynomial for outliers
            # split: boolean, default False
            #     split the data into years

            Returns
            -------
            Polynomial

            Performance Methods
            -------------------
            Additional methods to evaluate the process performance.

            Provide the values parameter to choose for which values the
            performance should be evaluated

            summary(values=None)
            control_chart(values=None)
            run_chart(values=None)
            process_capability(values=None)
            """
            # if split:
            #     return bluebelt.analysis.pattern.FramePolynomial(
            #         pd.Series(
            #             index=pd.MultiIndex.from_frame(self._obj.index.isocalendar()),
            #             data=self._obj.values,
            #         ).unstack(level=0),
            #         shape=shape,
            #         validation=validation,
            #         threshold=threshold,
            #         confidence=confidence,
            #         outlier_sigma=outlier_sigma,
            #         adjust=adjust,
            #     )
            if isinstance(self._obj, pd.DataFrame):
                pattern = pd.DataFrame(index=self._obj.index)
                residuals = pd.DataFrame(index=self._obj.index)
                outliers = pd.DataFrame(index=self._obj.index)
                upper = pd.DataFrame(index=self._obj.index)
                lower = pd.DataFrame(index=self._obj.index)
                out_of_bounds = pd.DataFrame(index=self._obj.index)
                within_bounds = pd.DataFrame(index=self._obj.index)
                
                statistic = pd.Series(dtype=float)
                p_value = pd.Series(dtype=float)
                rsq = pd.Series(dtype=float)
                std = pd.Series(dtype=float)
                outliers_count = pd.Series(dtype=int)
                sigma_level = pd.Series(dtype=float)
                bounds =pd.Series(dtype=float)


                for column in self._obj:
                    result = bluebelt.analysis.pattern.Polynomial(
                        self._obj[column],
                        shape=shape,
                        validation=validation,
                        threshold=threshold,
                        confidence=confidence,
                        outlier_sigma=outlier_sigma,
                        adjust=adjust,
                    )
                    pattern[column] = result.pattern
                    residuals[column] = result.residuals
                    outliers[column] = result.outliers
                    upper[column] = result.upper
                    lower[column] = result.lower
                    out_of_bounds[column] = result.out_of_bounds
                    within_bounds[column] = result.within_bounds

                    statistic = pd.concat([statistic, pd.Series({column: result.statistic})])
                    p_value = pd.concat([p_value, pd.Series({column: result.p_value})])
                    rsq = pd.concat([rsq, pd.Series({column: result.rsq})])
                    std = pd.concat([std, pd.Series({column: result.std})])
                    outliers_count = pd.concat([outliers_count, pd.Series({column: result.outliers_count})])
                    sigma_level = pd.concat([sigma_level, pd.Series({column: result.sigma_level})])
                    bounds = pd.concat([bounds, pd.Series({column: result.bounds})])
                    
                # REMOVE THIS LINE -- self refers to pattern class, not workload class

                self.pattern = pattern
                self.residuals = residuals
                self.outliers = outliers
                self.upper = upper
                self.lower = lower
                self.out_of_bounds = out_of_bounds
                self.within_bounds = within_bounds

                self.statistic = statistic
                self.p_value = p_value
                self.rsq = rsq
                self.std = std
                self.outliers_count = outliers_count
                self.sigma_level = sigma_level
                self.bounds = bounds
                
            else:
                return bluebelt.analysis.pattern.Polynomial(
                    self._obj,
                    shape=shape,
                    validation=validation,
                    threshold=threshold,
                    confidence=confidence,
                    outlier_sigma=outlier_sigma,
                    adjust=adjust,
                )

        def periodical(
            self,
            rule="1W",
            how="mean",
            confidence=0.8,
            outlier_sigma=2,
            adjust=True,
            **kwargs,
        ):

            """
            Find the periodical of an object.

            Parameters
            ----------
            series: pandas.Series or DataFrame
            rule: str, default "1W"
                period representation used for resampling the series
            how: str, default "mean"
                define how the period must be evaluated
                options are "mean", "min", "max" and "std"
            confidence: float, default 0.8
                the bandwidth confidence
            outlier_sigma: float, default 2
                outliers are datapoints outside the outlier_sigma fraction

            Returns
            -------
            Periodical

            Performance Methods
            -------------------
            Additional methods to evaluate the process performance.

            Provide the values parameter to choose for which values the
            performance should be evaluated

            summary(values=None)
            control_chart(values=None)
            run_chart(values=None)
            process_capability(values=None)
            """

            # TO DO | add dataframe support
            if isinstance(self._obj, pd.DataFrame):
                pattern = pd.DataFrame(index=self._obj.index)
                residuals = pd.DataFrame(index=self._obj.index)
                outliers = pd.DataFrame(index=self._obj.index)
                upper = pd.DataFrame(index=self._obj.index)
                lower = pd.DataFrame(index=self._obj.index)
                out_of_bounds = pd.DataFrame(index=self._obj.index)
                within_bounds = pd.DataFrame(index=self._obj.index)
                
                statistic = pd.Series(dtype=float)
                p_value = pd.Series(dtype=float)
                rsq = pd.Series(dtype=float)
                outliers_count = pd.Series(dtype=int)
                sigma_level = pd.Series(dtype=float)
                bounds =pd.Series(dtype=float)


                for column in self._obj:
                    result = bluebelt.analysis.pattern.Periodical(
                        self._obj[column],
                        rule=rule,
                        confidence=confidence,
                        outlier_sigma=outlier_sigma,
                        adjust=adjust,
                    )
                    pattern[column] = result.pattern
                    residuals[column] = result.residuals
                    outliers[column] = result.outliers
                    upper[column] = result.upper
                    lower[column] = result.lower
                    out_of_bounds[column] = result.out_of_bounds
                    within_bounds[column] = result.within_bounds

                    statistic = pd.concat([statistic, pd.Series({column: result.statistic})])
                    p_value = pd.concat([p_value, pd.Series({column: result.p_value})])
                    rsq = pd.concat([rsq, pd.Series({column: result.rsq})])
                    outliers_count = pd.concat([outliers_count, pd.Series({column: result.outliers_count})])
                    sigma_level = pd.concat([sigma_level, pd.Series({column: result.sigma_level})])
                    bounds = pd.concat([bounds, pd.Series({column: result.bounds})])
                    
                # REMOVE THIS LINE -- self refers to pattern class, not workload class

                self.pattern = pattern
                self.residuals = residuals
                self.outliers = outliers
                self.upper = upper
                self.lower = lower
                self.out_of_bounds = out_of_bounds
                self.within_bounds = within_bounds

                self.statistic = statistic
                self.p_value = p_value
                self.rsq = rsq
                self.outliers_count = outliers_count
                self.sigma_level = sigma_level
                self.bounds = bounds
            
            else:
                return bluebelt.analysis.pattern.Periodical(
                    self._obj,
                    rule=rule,
                    how=how,
                    confidence=confidence,
                    outlier_sigma=outlier_sigma,
                    adjust=adjust,
                    **kwargs,
                )

        def week_day(self, **kwargs):
            """
            Compare the distribution of data between week days

            Parameters
            ----------
            self : Series

            Returns
            -------
            WeekDay

            Attributes
            ----------
            series
                the transformed pandas.Series
            data
                the data
            equal_means
                the result of bluebelt.statistics.hypothesis_testing.EqualMeans().passed
            equal_variances
                the result of bluebelt.statistics.hypothesis_testing.EqualVariances().passed

            Methods
            -------
            plot: return the WeekDay plot
            """

            return bluebelt.analysis.datetime.WeekDay(self._obj, **kwargs)

        weekday = week_day

        def month_day(self, **kwargs):
            """
            Compare the distribution of data between month days

            Parameters
            ----------
            self : Series

            Returns
            -------
            MonthDay

            Attributes
            ----------
            series
                the transformed pandas.Series
            data
                the data
            equal_means
                the result of bluebelt.statistics.hypothesis_testing.EqualMeans().passed
            equal_variances
                the result of bluebelt.statistics.hypothesis_testing.EqualVariances().passed

            Methods
            -------
            plot: return the MonthDay plot
            """
            return bluebelt.analysis.datetime.MonthDay(self._obj, **kwargs)

        day = month_day

        def week(self, **kwargs):
            """
            Compare the distribution of data between weeks

            Parameters
            ----------
            self : Series

            Returns
            -------
            Week

            Attributes
            ----------
            series
                the transformed pandas.Series
            data
                the data
            equal_means
                the result of bluebelt.statistics.hypothesis_testing.EqualMeans().passed
            equal_variances
                the result of bluebelt.statistics.hypothesis_testing.EqualVariances().passed

            Methods
            -------
            plot: return the Week plot
            """
            return bluebelt.analysis.datetime.Week(self._obj, **kwargs)

        def month(self, **kwargs):
            """
            Compare the distribution of data between months

            Parameters
            ----------
            self : Series

            Returns
            -------
            Month

            Attributes
            ----------
            series
                the transformed pandas.Series
            data
                the data
            equal_means
                the result of bluebelt.statistics.hypothesis_testing.EqualMeans().passed
            equal_variances
                the result of bluebelt.statistics.hypothesis_testing.EqualVariances().passed

            Methods
            -------
            plot: return the Month plot
            """
            return bluebelt.analysis.datetime.Month(self._obj, **kwargs)

        def year(self, **kwargs):
            """
            Compare the distribution of data between years

            Parameters
            ----------
            self : Series

            Returns
            -------
            Year

            Attributes
            ----------
            series
                the transformed pandas.Series
            data
                the data
            equal_means
                the result of bluebelt.statistics.hypothesis_testing.EqualMeans().passed
            equal_variances
                the result of bluebelt.statistics.hypothesis_testing.EqualVariances().passed

            Methods
            -------
            plot: return the Year plot
            """
            return bluebelt.analysis.datetime.Year(self._obj, **kwargs)

    class planning:
        def __init__(self, pandas_obj):
            self._obj = pandas_obj

        def effort(self, skip_na=True, skip_zero=True):
            """
            Get the planning effort metrics.

            Parameters
            ----------
            self: Series or DataFrame

            Returns
            -------
            bluebelt.Effort

            Methods
            -------
            plot: return the WeekDay plot

            Performance Methods
            -------------------
            Additional methods to evaluate the process performance.

            Provide the values parameter to choose for which values the
            performance should be evaluated

            summary(values=None)
            control_chart(values=None)
            run_chart(values=None)
            process_capability(values=None)
            """
            return bluebelt.analysis.planning.Effort(
                self._obj, skip_na=skip_na, skip_zero=skip_zero
            )

        def ease(self, **kwargs):
            """
            Get the planning ease metrics.

            Parameters
            ----------
            self: Series or DataFrame

            Returns
            -------
            bluebelt.Ease

            Methods
            -------
            plot: return an ease plot

            Performance Methods
            -------------------
            Additional methods to evaluate the process performance.

            Provide the values parameter to choose for which values the
            performance should be evaluated

            summary(values=None)
            control_chart(values=None)
            run_chart(values=None)
            process_capability(values=None)
            """
            return bluebelt.analysis.planning.Ease(self._obj, **kwargs)

    class stats:
        def __init__(self, pandas_obj):
            self._obj = pandas_obj

        def index(self):
            return bluebelt.statistics.hypothesis_testing.index()

        # hypothesis testing
        def dagostino_pearson(self, alpha=0.05, **kwargs):
            return bluebelt.statistics.hypothesis_testing.DAgostinoPearson(
                self._obj, alpha=alpha
            )

        normal_distribution = dagostino_pearson

        def anderson_darling(self, dist="norm", alpha=0.05, **kwargs):
            return bluebelt.statistics.hypothesis_testing.AndersonDarling(
                self._obj, dist=dist, alpha=alpha
            )

        def std_within(self, how=None, observations=2, **kwargs):
            return bluebelt.statistics.std.StdWithin(
                self._obj, how=how, observations=observations, **kwargs
            )

        def one_sample_t(self, popmean=None, alpha=0.05, **kwargs):
            return bluebelt.statistics.hypothesis_testing.OneSampleT(
                self._obj, popmean=popmean, alpha=alpha, **kwargs
            )

        def wilcoxon(self, alpha=0.05, **kwargs):
            return bluebelt.statistics.hypothesis_testing.Wilcoxon(
                self._obj, alpha=alpha, **kwargs
            )

    class performance:
        def __init__(self, pandas_obj):
            self._obj = pandas_obj

        def summary(self, **kwargs):
            return bluebelt.analysis.performance.Summary(self._obj, **kwargs)

        def control_chart(self, **kwargs):
            """
            Calculate and display a control chart.

            Parameters
            ----------
            self : Series
            format_stats: str, default "1.2f"
                the formatting for all statistics returned

            Returns
            -------
            ControlChart

            Attributes
            ----------
            series
                the Series

            mean:
                float with the mean
            std:
                float with the standard deviation
            ucl:
                float with the upper control limit
            lcl:
                float with the lower control limit
            outliers:
                the outliers as a Series
            outlier_count:
                int with the number of outliers

            Methods
            -------
            plot:
                return the ControlChart plot

            """
            return bluebelt.analysis.performance.ControlChart(self._obj, **kwargs)

        def run_chart(self, **kwargs):
            """
            Calculate and display a run chart

            The number of runs about the median is the total number of runs above the median
            and the total number of runs below the median.
            A run about the median is one or more consecutive points on the same side of the
            center line.
            A run ends when the line that connects the points crosses the center line.
            A new run begins with the next plotted point.
            A data point equal to the median belongs to the run below the median.

            The number of runs up or down is the total count of upward and downward runs in
            the series.
            A run up or down ends when the direction changes.

            Clustering, mixtures, trends and oscillation
            A p-value that is less than the specified level of significance (alpha) indicates
            clustering, mixtures, trends and/or oscillation

            Parameters
            ----------
            self : Series
            alpha: float, default 0.05
                the threshold for clustering, mixtures, trends and oscillation
            format_stats: str, default "1.2f"
                the formatting for all statistics returned

            Returns
            -------
            RunChart

            Attributes
            ----------
            series
                the Series
            alpha:
            format_stats:
            metrics:
                prints the run chart metrics (all metrics below)
            runs_about:
                the number of runs about the median
            expected_runs_about:
                the expected number of runs about the median
            longest_run_about:
                the longest run about the median
            runs_up_or_down:
                the number of runs up or down
            expected_runs_up_or_down:
                the expected number of runs up or down
            longest_run_up_or_down:
                the longest run up or down
            p_value_clustering:
                the p-value for clustering
            p_value_mixtures:
                the p-value for mixtures
            p_value_trends:
                the p-value for trends
            p_value_oscillation:
                the p-value for oscillation
            clustering:
                boolean value for clustering
            mixtures:
                boolean value for mixtures
            trends:
                boolean value for trends
            oscillation:
                boolean value for oscillation
            longest_runs_about:
                a list of the longest run or runs about the median
            longest_runs_up_or_down:
                a list of the longest run or runs up or down

            Methods
            -------
            plot:
                return the RunChart plot

            """
            return bluebelt.analysis.performance.RunChart(self._obj, **kwargs)

        def process_capability(self, **kwargs):
            """
            Calculate and display the process capability

            Parameters
            ----------
            self : Series
            target: float, default None
                target value for the process
            usl: float, default None
                upper specification limit (usl and ub cannot be specified both)
            ub: float, default None
                upper bound (usl and ub cannot be specified both)
            lsl: float, default None
                lower specification limit
            lb: float, default None
                lower bound (lsl and lb cannot be specified both)
            tolerance: float, default 6.0
                sigma tolerance of the process


            alpha: float, default 0.05
                the threshold for clustering, mixtures, trends and oscillation
            format_stats: str, default "1.2f"
                the formatting for all statistics returned

            Returns
            -------
            RunChart

            Attributes
            ----------
            series
                the Series
            min
            max
            mean
            std
            std_within

            ### observed performance
            observed_lt_lsl
            observed_gt_usl
            observed_performance

            ### expected performance
            expected_lt_lsl_within
            expected_gt_usl_within
            expected_performance_within

            expected_lt_lsl_overall
            expected_gt_usl_overall
            expected_performance_overall

            ### within capability
            cp
            cpl
            cpu
            cpk
            ccpk

            ### overall capability
            pp
            ppl
            ppu
            ppk
            cpm

            Methods
            -------
            df
                show the process capability in a pandas.DataFrame
            plot
                return the ProcessCapability plot

            """
            return bluebelt.analysis.performance.ProcessCapability(self._obj, **kwargs)

    class forecast:
        def __init__(self, pandas_obj):
            self._obj = pandas_obj

        def MAE(self, **kwargs):
            """
            Find the Mean Absolute Error.

            Parameters
            ----------
            self: pandas.DataFrame
            forecast: str, default None
                the column name with the forecast data
                if forecast is None the first column of the DataFrame will be
                used as forecast data
            actuals: str, default None
                the column name with the actuals data
                if actuals is None the second column of the DataFrame will be
                used as actuals data

            Returns
            -------
            MAE

            Performance Methods
            -------------------
            Additional methods to evaluate the process performance.

            Provide the values parameter to choose for which values the
            performance should be evaluated

            summary(values=None)
            control_chart(values=None)
            run_chart(values=None)
            process_capability(values=None)

            """
            return bluebelt.analysis.forecast.MAE(self._obj, **kwargs)

        def MAPE(self, **kwargs):
            """
            Find the Mean Absolute Percentage Error.

            Parameters
            ----------
            self: pandas.DataFrame
            forecast: str, default None
                the column name with the forecast data
                if forecast is None the first column of the DataFrame will be
                used as forecast data
            actuals: str, default None
                the column name with the actuals data
                if actuals is None the second column of the DataFrame will be
                used as actuals data

            Returns
            -------
            MAPE

            Performance Methods
            -------------------
            Additional methods to evaluate the process performance.

            Provide the values parameter to choose for which values the
            performance should be evaluated

            summary(values=None)
            control_chart(values=None)
            run_chart(values=None)
            process_capability(values=None)

            """
            return bluebelt.analysis.forecast.MAPE(self._obj, **kwargs)

        def SMAPE(self, **kwargs):
            """
            Find the Symmetrical Mean Absolute Percentage Error.

            Parameters
            ----------
            self: pandas.DataFrame
            forecast: str, default None
                the column name with the forecast data
                if forecast is None the first column of the DataFrame will be
                used as forecast data
            actuals: str, default None
                the column name with the actuals data
                if actuals is None the second column of the DataFrame will be
                used as actuals data
            adjust: boolean, default True
                calculate use the adjusted SMAPE

            Returns
            -------
            SMAPE

            Performance Methods
            -------------------
            Additional methods to evaluate the process performance.

            Provide the values parameter to choose for which values the
            performance should be evaluated

            summary(values=None)
            control_chart(values=None)
            run_chart(values=None)
            process_capability(values=None)

            """
            return bluebelt.analysis.forecast.SMAPE(self._obj, **kwargs)

        def MDA(self, **kwargs):
            """
            Find the Mean Directional Accuracy.

            Parameters
            ----------
            self: pandas.DataFrame
            forecast: str, default None
                the column name with the forecast data
                if forecast is None the first column of the DataFrame will be
                used as forecast data
            actuals: str, default None
                the column name with the actuals data
                if actuals is None the second column of the DataFrame will be
                used as actuals data

            Returns
            -------
            MDA
            """
            return bluebelt.analysis.forecast.MDA(self._obj, **kwargs)

        def MPE(self, **kwargs):
            """
            Find the Mean Percentage Error.

            Parameters
            ----------
            self: pandas.DataFrame
            forecast: str, default None
                the column name with the forecast data
                if forecast is None the first column of the DataFrame will be
                used as forecast data
            actuals: str, default None
                the column name with the actuals data
                if actuals is None the second column of the DataFrame will be
                used as actuals data
            confidence: float, default 0.95
                the confidence for measuring the confidence interval of the mean

            Returns
            -------
            MPE

            Performance Methods
            -------------------
            Additional methods to evaluate the process performance.

            Provide the values parameter to choose for which values the
            performance should be evaluated

            summary(values=None)
            control_chart(values=None)
            run_chart(values=None)
            process_capability(values=None)

            """
            return bluebelt.analysis.forecast.MPE(self._obj, **kwargs)

        def NFM(self, **kwargs):
            """
            Find the Normalized Forecast Metric. If the result > 0 the forecast
            tends to be greater than the actuals.

            Parameters
            ----------
            self: pandas.DataFrame
            forecast: str, default None
                the column name with the forecast data
                if forecast is None the first column of the DataFrame will be
                used as forecast data
            actuals: str, default None
                the column name with the actuals data
                if actuals is None the second column of the DataFrame will be
                used as actuals data
            confidence: float, default 0.95
                the confidence for measuring the confidence interval of the mean

            Returns
            -------
            NFM

            """
            return bluebelt.analysis.forecast.NFM(self._obj, **kwargs)

        mae = MAE
        mape = MAPE
        smape = SMAPE
        mda = MDA
        mpe = MPE
        nfm = NFM
        bias = NFM

    # graphs
    
    def line(
        self,
        xlim=(None, None),
        ylim=(None, None),
        max_xticks=None,
        format_xticks=None,
        format_yticks=None,
        title=None,
        xlabel=None,
        ylabel=None,
        legend=True,
        path=None,
        **kwargs,
    ):
        return bluebelt.graph.graph.line(
            self._obj,
            xlim=xlim,
            ylim=ylim,
            max_xticks=max_xticks,
            format_xticks=format_xticks,
            format_yticks=format_yticks,
            title=title,
            xlabel=xlabel,
            ylabel=ylabel,
            legend=legend,
            path=path,
            **kwargs,
        )

    def bar(
        self,
        xlim=(None, None),
        ylim=(None, None),
        max_xticks=None,
        format_xticks=None,
        format_yticks=None,
        title=None,
        xlabel=None,
        ylabel=None,
        legend=True,
        path=None,
        **kwargs,
    ):
        return bluebelt.graph.graph.bar(
            self._obj,
            xlim=xlim,
            ylim=ylim,
            max_xticks=max_xticks,
            format_xticks=format_xticks,
            format_yticks=format_yticks,
            title=title,
            xlabel=xlabel,
            ylabel=ylabel,
            legend=legend,
            path=path,
            **kwargs,
        )

    def area(
        self,
        xlim=(None, None),
        ylim=(None, None),
        max_xticks=None,
        format_xticks=None,
        format_yticks=None,
        title=None,
        xlabel=None,
        ylabel=None,
        legend=True,
        path=None,
        **kwargs,
    ):
        return bluebelt.graph.graph.area(
            self._obj,
            xlim=xlim,
            ylim=ylim,
            max_xticks=max_xticks,
            format_xticks=format_xticks,
            format_yticks=format_yticks,
            title=title,
            xlabel=xlabel,
            ylabel=ylabel,
            legend=legend,
            path=path,
            **kwargs,
        )

    def hist(
        self,
        bins=20,
        xlim=(None, None),
        ylim=(None, None),
        format_xticks=None,
        format_yticks=None,
        title=None,
        xlabel=None,
        ylabel=None,
        legend=True,
        path=None,
        **kwargs,
    ):
        return bluebelt.graph.graph.hist(
            self._obj,
            bins=bins,
            xlim=xlim,
            ylim=ylim,
            format_xticks=format_xticks,
            format_yticks=format_yticks,
            title=title,
            xlabel=xlabel,
            ylabel=ylabel,
            legend=legend,
            path=path,
            **kwargs,
        )

    histogram = hist

    def box(
        self,
        xlim=(None, None),
        ylim=(None, None),
        format_xticks=None,
        format_yticks=None,
        title=None,
        xlabel=None,
        ylabel=None,
        path=None,
        **kwargs,
    ):
        return bluebelt.graph.graph.box(
            self._obj,
            xlim=xlim,
            ylim=ylim,
            format_xticks=format_xticks,
            format_yticks=format_yticks,
            title=title,
            xlabel=xlabel,
            ylabel=ylabel,
            path=path,
            **kwargs,
        )

    boxplot = box

    def waterfall(
        self,
        horizontal=False,
        swapx=False,
        swapy=False,
        width=0.6,
        height=0.6,
        xlim=(None, None),
        ylim=(None, None),
        format_xticks=None,
        format_yticks=None,
        title=None,
        xlabel=None,
        ylabel=None,
        path=None,
        **kwargs,
    ):
        return bluebelt.graph.graph.waterfall(
            self._obj,
            horizontal=horizontal,
            swapx=swapx,
            swapy=swapy,
            width=width,
            height=height,
            xlim=xlim,
            ylim=ylim,
            format_xticks=format_xticks,
            format_yticks=format_yticks,
            title=title,
            xlabel=xlabel,
            ylabel=ylabel,
            path=path,
            **kwargs,
        )



import numpy as np
import copy

import bluebelt

import bluebelt.helpers.check as check


def _filter(cls):
    def weekend(
        self, drop=None, fill_value=np.nan, inplace=None, weekend_days=[5, 6]
    ):
        """
        Fill all non-weekend data with the fill_value. If fill_value=None the
        Series or DataFrame will contain only the weekend days.

        The Series or DataFrame must have a DatetimeIndex

        Parameters
        ----------
        self : Series or DataFrame
        fill_value: object or None, default numpy.nan
        drop: boolean, default None
            if True the filtered values will be dropped
            if drop is None global settings will be applied, default True
        inplace: boolean, default None
            if True the function will be on the object itself
            if False the function will return a new object
            if inplace is None global settings will be applied, default True
        weekend_days: list of int, default [5, 6]

        Returns
        -------
        Series or DataFrame
        """
        # settings
        drop = drop if drop is not None else bluebelt.config("drop")
        inplace = inplace if inplace is not None else bluebelt.config("inplace")
        
        # checks
        check.has_datetimeindex(self._obj)
        series = copy.deepcopy(self._obj) if not inplace else self._obj

        
        if fill_value is not None and not drop:
            series.loc[~series.index.weekday.isin(weekend_days)] = fill_value
        else:
            series = series[series.index.weekday.isin(weekend_days)]

        # only if Workload Class is used directly modify _obj
        # if self.__class__.__name__ == "Workload":
        #     self._obj = series
        
        return series


    setattr(cls, "weekend", weekend)

    return cls


# 

class WorkforceGroup:
    # e.g.
    # Full-Time Employees.
    # Part-Time Employees.
    # Seasonal Employees.
    # Temporary Employees.
    
    def __init__(self, name, count, hours, hours_min, hours_max, period, **kwargs):

        self.name = name
        self.count = count
        self.hours = hours
        self.hours_min = hours_min
        self.hours_max = hours_max
        self.period = period

    def __repr__(self):
        return f'{self.__class__.__name__}(name={self.name}, count={self.count}, hours={self.hours}, hours_min={self.hours_min}, hours_max={self.hours_max}, period={self.period})'
        

class Workforce:
    def __init__(self, types, **kwargs):

        self.types = types


from multiprocessing.sharedctypes import Value
import numpy as np
import pandas as pd
import datetime
import math

import bluebelt.helpers.check as check


def resolution_methods(cls):
    def sum(self):

        result = self.grouped.sum(min_count=1)
        if isinstance(self, Flatten):
            result = result.reindex_like(self._obj, method="ffill")
        return result

    def mean(self):
        result = self.grouped.mean()
        if isinstance(self, Flatten):
            result = result.reindex_like(self._obj, method="ffill")
        return result

    def var(self):
        result = self.grouped.var()
        if isinstance(self, Flatten):
            result = result.reindex_like(self._obj, method="ffill")
        return result

    def std(self):
        result = self.grouped.std()
        if isinstance(self, Flatten):
            result = result.reindex_like(self._obj, method="ffill")
        return result

    def min(self):
        result = self.grouped.min()
        if isinstance(self, Flatten):
            result = result.reindex_like(self._obj, method="ffill")
        return result

    def max(self):
        result = self.grouped.max()
        if isinstance(self, Flatten):
            result = result.reindex_like(self._obj, method="ffill")
        return result

    def count(self):
        result = self.grouped.count()
        if isinstance(self, Flatten):
            result = result.reindex_like(self._obj, method="ffill")
        return result

    def value_range(self):
        result = self.grouped.apply(lambda x: x.max() - x.min())
        if isinstance(self, Flatten):
            result = result.reindex_like(self._obj, method="ffill")
        return result

    def subseries(self, count, size=1):
        """
        Count the number of times a list of 'count' items with size 'size' fit
        in each group.

        Parameters
        ----------
        self : GroupBy or Resampling object
        count: int
            The length of the sub-series.
        size: int
            The size of each element in the sub-series.

        Returns
        -------
        Series

        Example
        -------

        GroupBy or Resampling object: pd.Series([10, 8, 3, 3, 5])
        count: 3
        size: 1

        frame.blue.resample(rule=7).subseries(3, 1)
        >>> 9

        Every step subtracts (1, 1, 1, 0, 0), (1, 1, 0, 1, 0), (1, 1, 0, 0, 1),
        (1, 0, 0, 1, 1) or (0, 0, 1, 1, 1) from the group.

        step 0: (3, 3, 5, 8, 10)
        step 1: (3, 3, 4, 7, 9)
        step 2: (3, 3, 3, 6, 8)
        step 3: (2, 3, 3, 5, 7)
        step 4: (2, 2, 3, 4, 6)
        step 5: (2, 2, 2, 3, 5)
        step 6: (1, 2, 2, 2, 4)
        step 7: (1, 1, 1, 2, 3)
        step 8: (0, 1, 1, 1, 2)
        step 9: (0, 0, 0, 1, 1)
        """

        if not isinstance(count, int):
            raise ValueError("count must be an int")
        if not isinstance(size, (float, int)):
            raise ValueError("size must be float or int")
        # if isinstance(self._obj, pd.DataFrame):
        #     data = self._obj.sum(axis=1, numeric_only=True)

        def _subseries_count(series, count=3, size=1):
            # helper for resample.subseries
            series = pd.Series(series) / size
            result = series.sum() * count
            for i in range(count, 0, -1):
                result = np.minimum(
                    result,
                    math.floor(series.nsmallest(len(series) - count + i).sum() / i),
                )
            return result

        if isinstance(self._obj, pd.Series):
            result = pd.Series(
                index=self.grouped.groups.keys(),
                data=self.grouped.apply(
                    lambda x: _subseries_count(series=x, count=count, size=size)
                ).values,
                name=f"subseries {str(count)} of {str(size)}",
            )
        elif isinstance(self._obj, pd.DataFrame):
            result = pd.DataFrame(
                index=self.grouped.groups.keys(),
                data=self.grouped.apply(
                    lambda x: _subseries_count(series=x, count=count, size=size),
                ).values,
                columns=self._obj.columns,
            )

        if isinstance(self, Flatten):
            result = result.reindex_like(self._obj, method="ffill")
        return result

    def diff_quantity(self, shift=1, skip_na=True, skip_zero=True):
        # if this week > last week: this week - last week / this week
        # if this week < last week: last week - this week / last week

        if isinstance(self._obj, pd.DataFrame):
            data = self._obj.sum(
                axis=1, min_count=1
            )  # min_count=1 prevents np.nan values to become zeros
        else:
            data = self._obj

        # resample the data and transpose into a DataFrame with weekdays as columns
        data = data.resample(rule=self.rule, origin=self.origin, **self.kwargs)
        data = data.apply(
            lambda x: pd.DataFrame(
                index=x.index.isocalendar().day,
                data=x.values,
            ).transpose()
        )

        # drop the extra index level that was created in the transpose
        data = data.droplevel(-1)

        # get the difference, but only for days where both weeks have a proper value
        if skip_na and skip_zero:
            result = (
                data[(data.shift(shift).notna()) & (data.shift(shift) != 0)]
                - data.shift(shift)[(data.notna()) & (data != 0)]
            ).sum(axis=1).abs() / np.maximum(
                data[(data.shift(shift).notna()) & (data.shift(shift) != 0)].sum(axis=1),
                data.shift(shift)[(data.notna()) & (data != 0)].sum(axis=1),
            )
        elif skip_na:
            result = (data[data.shift(shift).notna()] - data.shift(shift)[data.notna()]).sum(
                axis=1
            ).abs() / np.maximum(
                data[data.shift(shift).notna()].sum(axis=1),
                data.shift(shift)[data.notna()].sum(axis=1),
            )
        elif skip_zero:
            result = (data[data.shift(shift) != 0] - data.shift(shift)[data != 0]).sum(
                axis=1
            ).abs() / np.maximum(
                data[data.shift(shift) != 0].sum(axis=1), data.shift(shift)[data != 0].sum(axis=1)
            )
        else:
            result = (data - data.shift(shift)).sum(axis=1).abs() / np.maximum(
                data.sum(axis=1), data.shift(shift).sum(axis=1)
            )

        return result

    def diff_distribution(self, shift=1, skip_na=True, skip_zero=True):
        # only works with resample('1w') or equivalent
        if self.rule != datetime.timedelta(days=7):
            raise ValueError(
                f"The object must be resampled to weekly data not {self.rule}"
            )

        # if a DataFrame is passed then sum the columns
        if isinstance(self._obj, pd.DataFrame):
            data = self._obj.sum(axis=1, numeric_only=True)
        else:
            data = self._obj

        # resample the data and transpose into a DataFrame with weekdays as columns
        data = data.resample(rule=self.rule, origin=self.origin, **self.kwargs)
        data = data.apply(
            lambda x: pd.DataFrame(
                index=x.index.isocalendar().day,
                data=x.values,
            ).transpose()
        )

        # drop the extra index level that was created in the transpose
        data = data.droplevel(-1)

        # get diff_quantity to calculate the relative difference
        if skip_na and skip_zero:
            diff = (
                (
                    data[(data.shift(shift).notna()) & (data.shift(shift) != 0)]
                    - data.shift(shift)[(data.notna()) & (data != 0)].multiply(
                        data[(data.shift(shift).notna()) & (data.shift(shift) != 0)].sum(axis=1)
                        / data.shift(shift)[(data.notna()) & (data != 0)].sum(axis=1),
                        axis=0,
                    )
                )
                .abs()
                .sum(axis=1, min_count=1)
            )
        elif skip_na:
            diff = (
                (
                    data[data.shift(shift).notna()]
                    - data.shift(shift)[data.notna()].multiply(
                        data[data.shift(shift).notna()].sum(axis=1)
                        / data.shift(shift)[data.notna()].sum(axis=1),
                        axis=0,
                    )
                )
                .abs()
                .sum(axis=1, min_count=1)
            )
        elif skip_zero:
            diff = (
                (
                    data[data.shift(shift) != 0]
                    - data.shift(shift)[data != 0].multiply(
                        data[data.shift(shift) != 0].sum(axis=1)
                        / data.shift(shift)[data != 0].sum(axis=1),
                        axis=0,
                    )
                )
                .abs()
                .sum(axis=1, min_count=1)
            )
        else:
            diff = (
                (
                    data
                    - data.shift(shift).multiply(
                        data.sum(axis=1) / data.shift(shift).sum(axis=1),
                        axis=0,
                    )
                )
                .abs()
                .sum(axis=1, min_count=1)
            )

        # build result
        result = pd.Series(
            diff / (data.sum(axis=1, min_count=1) * 2),
            name="distribution",
        )

        # and maybe flatten
        if isinstance(self, Flatten):
            result = result.reindex_like(self._obj, method="ffill")

        return result

    def diff_skills(self, shift=1, skip_na=True, skip_zero=True):

        """
        Calculate the difference in skills distribution from week to week in
        from a resampled DataFrame.

        Parameters
        ----------
        self : GroupBy or Resampling object

        Returns
        -------
        Series

        Example
        -------
        index = pd.date_range(start='2022-03-07 00:00:00', periods=21)
        data = {'skill 1': [20,20,10,15,10,10,10,
                            20,20,10,10,10,10,10,
                            20,20,10,10,10,10,10,],
                'skill 2': [20,10,20,5,10,10,10,
                            20,20,10,10,10,10,10,
                            20,10,20,10,10,10,10,],
                }
        frame = pd.DataFrame(index=index, data=data)

        >>> frame.blue.resample(rule=7).diff_skills()

        2022-03-07         NaN
        2022-03-14    0.027778
        2022-03-21    0.000000
        Name: skills, dtype: float64
        """

        # only works with resample('1w') or equivalent
        if self.rule != datetime.timedelta(days=7):
            raise ValueError(
                f"The object must be resampled to weekly data not {self.rule}"
            )

        # if the object is a Series return Series of zeros.
        if isinstance(self._obj, pd.Series):
            return pd.Series(
                index=self.grouped.sum().index,
                data=np.zeros(self.grouped.sum().size),
                name="skills",
            )

        # resample the data and transpose into a DataFrame with weekdays as columns
        data = self.grouped.apply(
            lambda x: pd.DataFrame(
                index=x.index.isocalendar().day,
                data=x.values,
                columns=x.columns,
            ).transpose()
        )

        # build a filter for the dataframe for weekdays with value zero or np.nan
        summed = data.groupby(level=0).apply(pd.DataFrame.sum, skipna=True)

        filt_index = pd.MultiIndex.from_product([summed.index, self._obj.columns])
        filt_data = np.repeat(summed.values, repeats=self._obj.shape[1], axis=0)

        if skip_na and skip_zero:
            filt_data = (filt_data != 0) & (np.logical_not(np.isnan(filt_data)))
        elif skip_na:
            filt_data = np.logical_not(np.isnan(filt_data))
        elif skip_zero:
            filt_data = filt_data != 0
        else:
            filt_data = np.full(filt_data.shape, True)

        filt = pd.DataFrame(
            index=filt_index,
            data=filt_data,
            columns=summed.columns,
        )

        # get diff_quantity to calculate the relative difference
        if skip_na and skip_zero:
            diff_quantity = summed[
                (summed.shift(shift).notna()) & (summed.shift(shift) != 0)
            ].sum(axis=1) / summed.shift(shift)[(summed.notna()) & (summed != 0)].sum(axis=1)
        elif skip_na:
            diff_quantity = summed[summed.shift(shift).notna()].sum(axis=1) / summed.shift(shift)[
                summed.notna()
            ].sum(axis=1)
        elif skip_zero:
            diff_quantity = summed[summed.shift(shift) != 0].sum(axis=1) / summed.shift(shift)[
                summed != 0
            ].sum(axis=1)
        else:
            diff_quantity = summed.sum(axis=1) / summed.shift(shift).sum(axis=1)

        # calculate diff_skills

        # filter the data with shifted filter and the shifted data with the unshifted filter
        # remember that the shift is equal to the number of skills (self._obj.shape[1])
        # sum the data for the weekdays
        # unstack the skills => the result is grouped data with skills as columns

        diff = (
            # the data with shifted filter minus the shifted data with filter
            (
                data[filt.shift(self._obj.shape[1] * shift)].sum(axis=1).unstack(level=1)
                - data.shift(self._obj.shape[1] * shift)[filt]
                .sum(axis=1)
                .unstack(level=1)
                .multiply(diff_quantity, axis=0)
            )
            .abs()
            .sum(axis=1, min_count=1)
        )

        # get the difference relative to the sum of the row with the shifted filter applied
        result = pd.Series(
            diff
            / (
                data[filt.shift(self._obj.shape[1] * shift)]
                .sum(axis=1)
                .unstack(level=1)
                .sum(axis=1, min_count=1)
                * 2
            ),
            name="skills",
        )

        if isinstance(self, Flatten):
            result = result.reindex_like(self._obj, method="ffill")
        return result

    setattr(cls, "sum", sum)
    setattr(cls, "mean", mean)
    setattr(cls, "var", var)
    setattr(cls, "std", std)
    setattr(cls, "min", min)
    setattr(cls, "max", max)
    setattr(cls, "count", count)
    setattr(cls, "value_range", value_range)
    setattr(cls, "subseries", subseries)
    setattr(cls, "diff_quantity", diff_quantity)
    setattr(cls, "diff_distribution", diff_distribution)
    setattr(cls, "diff_skills", diff_skills)
    return cls


@resolution_methods
class Resample:
    def __init__(self, _obj, rule, **kwargs):
        self._obj = _obj
        self.rule = _get_rule(rule)
        self.origin = _get_origin(_obj, self.rule)
        self.kwargs = kwargs
        self.calculate()

    def calculate(self, **kwargs):
        self.grouped = self._obj.resample(
            rule=self.rule, origin=self.origin, **self.kwargs
        )

    def __repr__(self):
        return self.grouped.__repr__()


@resolution_methods
class Flatten:
    """
    It is just a copy of Resample
    """

    def __init__(self, _obj, rule, **kwargs):
        self._obj = _obj
        self.rule = _get_rule(rule)
        self.origin = _get_origin(_obj, self.rule)
        self.kwargs = kwargs
        self.calculate()

    def calculate(self, **kwargs):
        self.grouped = self._obj.resample(
            rule=self.rule, origin=self.origin, **self.kwargs
        )

    def __repr__(self):
        return self.grouped.__repr__()


def _get_origin(_obj, rule):
    if (rule / datetime.timedelta(days=7)).is_integer():
        days = int(
            ((_obj.index.isocalendar().week[0] - 1) * 7) + _obj.index[0].weekday()
        )
        return _obj.index[0] - datetime.timedelta(days=days)
    elif (rule / datetime.timedelta(days=1)).is_integer():
        days = int(_obj.index[0].weekday())
        return _obj.index[0] - datetime.timedelta(days=days)
    else:
        return "start_day"


def _get_rule(rule):
    if isinstance(rule, str):
        rule = rule.replace(" ", "").lower()
        unit = rule.lstrip("0123456789")
        count = int(rule[0 : len(rule) - len(unit)]) if len(rule) > len(unit) else 1
        if unit in {"μs", "microsecond", "microseconds"}:
            return datetime.timedelta(microseconds=count)
        if unit in {"ms", "millisecond", "milliseconds"}:
            return datetime.timedelta(milliseconds=count)
        if unit in {"s", "sec", "second", "seconds"}:
            return datetime.timedelta(seconds=count)
        if unit in {"m", "min", "minute", "minutes"}:
            return datetime.timedelta(minutes=count)
        if unit in {"h", "hr", "hrs", "hour", "hours"}:
            return datetime.timedelta(hours=count)
        if unit in {"d", "day", "days"}:
            return datetime.timedelta(days=count)
        if unit in {"w", "wk", "week", "weeks"}:
            return datetime.timedelta(weeks=count)
    elif isinstance(rule, int):
        return datetime.timedelta(days=rule)
    else:
        return rule



import numpy as np
import datetime

def last_iso_week_in_year(year):
    return datetime.date(year, 12, 28).isocalendar()[1]

def year_with_most_iso_weeks(years):
    return years[np.argmax([last_iso_week_in_year(year) for year in years if isinstance(year, (np.uint32, int))])]


import numpy as np
import pandas as pd

import bluebelt.helpers.constants as constants

class StdWithin():
    """
    Calculate the standard deviation within subgroups.
        arguments
        frame: pandas.DataFrame or pandas.Series
        columns: list
            a list of column names
            if not provided all columns will be treated as subgroups
            if frame is a Pandas Series only the first column will be used
            default value: None
        how: str
            the method of estimating std_within
            options when subgroup size == 1 (default when frame is a Pandas Series):
            - 'amr' or 'average_moving_range'
                optional argument: observations=2
            - 'mmr' or 'median_moving_range'
                optional argument: observations=2
            - 'mssd' or 'sqrt_of_mssd'
            default value: 'amr'

            options when subgroup size > 1:
            - 'pooled_std'
            - 'r_bar'
            - 's_bar'
            default value: 'pooled_std'

        observations: int
            used for average moving range and median moving range estimation
            default value: 2
    """

    def __init__(self, frame, axis=0, columns=None, how=None, observations=2, **kwargs):

        if isinstance(frame, pd.DataFrame) and isinstance(columns, (str, list)):
            frame = frame[columns]

        if isinstance(frame, pd.DataFrame) and axis == 0:
            self.data = frame.values
        elif isinstance(frame, pd.DataFrame):
            self.data = frame.T.values
        elif isinstance(frame, pd.Series):
            self.data = frame.values

        self.frame = frame
        self.columns = columns
        self.axis = axis
        self.how = how
        self.observations = observations
        self.nrows = self.data.shape[0]
        self.ncols = 1 if self.data.ndim == 1 else self.data.shape[1]
        self.calculate(**kwargs)

    def calculate(self, **kwargs):
        # choose how depending on subgroup size
        if self.ncols == 1:
            if self.how in ['mssd', 'sqrt_of_mssd']:
                self.how = 'mssd'
                self.std = _sqrt_of_mssd(self, **kwargs)
            elif self.how in ['mmr', 'median_moving_range']:
                self.how = 'mmr'
                self.std = _median_moving_range(self, **kwargs)
            else:
                self.how = 'amr'
                self.std = _average_moving_range(self, **kwargs)
        
        else:
            if self.how=='r_bar':
                self.std = _r_bar(self, **kwargs)
            elif self.how == 's_bar':
                self.std = _s_bar(self, **kwargs)
            else:
                self.how = 'pooled std'
                self.std = _pooled_std(self, **kwargs)

    def __str__(self):
        return ""
    
    def __repr__(self):
        return (f'{self.__class__.__name__}(nrows={self.nrows}, ncols={self.ncols}, how=\'{self.how}\', axis={self.axis}, std={self.std:1.4f})')
    

def std_within(frame, columns=None, axis=0, how=None, observations=2, **kwargs):
    StdWithin(frame, columns=columns, axis=axis, how=how, observations=observations, **kwargs)  

# standard deviation within subgroups with subgroup size == 1
def _average_moving_range(std_obj, **kwargs):
    """
    Calculate the average moving range
    """
        
    periods = std_obj.observations - 1
    if std_obj.observations >= std_obj.nrows:
        raise ValueError(f'The number of observations ({std_obj.observations}) must be lower then the length of the data({std_obj.nrows})')

    return pd.Series(std_obj.data).diff(periods=periods).apply(abs)[periods:].dropna().mean() / constants.d2(std_obj.observations)

def _median_moving_range(std_obj, **kwargs):
    """
    Calculate the median moving range
    """
    
    periods = std_obj.observations - 1
    if std_obj.observations >= std_obj.nrows:
        raise ValueError(f'The number of observations ({std_obj.observations}) must be lower then the length of the series({frame.shape[0]})')

    return pd.Series(std_obj.data).diff(periods=periods).apply(abs)[periods:].dropna().median() / constants.d2(std_obj.observations)

def _sqrt_of_mssd(std_obj, **kwargs):
    """
    Calculate the square root of half of the mean of the squared successive differences
    """
    
    return (0.5 * sum(pd.Series(std_obj.data).diff(periods=1).dropna()**2) / (std_obj.nrows - 1) )**0.5 / constants.c4_(std_obj.nrows)

 
# standard deviation within subgroups with subgroup size > 1
def _pooled_std(std_obj, **kwargs):
    """
    Calculate pooled standard deviation for pandas DataFrame columns.
    """

    # check if subgroup size is constant
    if len(set([x[~np.isnan(x)].size for x in std_obj.data])) == 1:
        # all subgroups have the same size
        # S_p = (sum([frame[col].var(skipna=True, ddof=1) for col in frame.columns]) / len(frame.columns)) ** 0.5
        # d = sum([frame[col].dropna(axis=0).size for col in frame.columns]) - len(frame.columns) + 1
        S_p = (sum([x[~np.isnan(x)].var(ddof=1) for x in std_obj.data]) / std_obj.nrows) ** 0.5
        d = sum([x[~np.isnan(x)].size for x in std_obj.data]) - std_obj.nrows + 1
        
    else:
        # calculate overall mean
        arr = std_obj.data.flatten()
        Xi_bar = arr[~np.isnan(arr)].mean()
        
        # S_p = (sum([sum((frame[col].dropna(axis=0)-Xi_bar)**2) for col in frame.columns]) / sum([(frame[col].dropna(axis=0).size - 1) for col in frame.columns])) ** 0.5
        # d = sum([(frame[col].dropna(axis=0).size - 1) for col in frame.columns]) + 1

        S_p = (sum([sum((x[~np.isnan(x)]-Xi_bar)**2) for x in std_obj.data]) / sum([(x[~np.isnan(x)].size - 1) for x in std_obj.data]) ) ** 0.5
        d = sum([(x[~np.isnan(x)].size - 1) for x in std_obj.data]) + 1

    return S_p / constants.c4(d)

def _r_bar(std_obj, columns=None, **kwargs):
    """
    Calculate r_bar; Average of subgroup ranges for pandas DataFrame columns.
    r_bar is used for estimating std_within when subgroup size > 1
        arguments
        frame:      Pandas Dataframe
        columns:    list of colums; if not provided all columns will be treated as subgroups

    std_within = sum_i ( (f_i * r_i) / (d2(n_i)) ) / sum_i (f_i)
    f_i = ( d2(n_i)**2 ) / ( d3(n_i)**2 )

    If all n_i are the same (all subgroup sizes are of equal length) then std_within is just the adjusted average of subgroup ranges.
    std_within = r_bar / d2(n_i)
    """
    
    # setup lambda functions
    subgroup_range = lambda x: (x.max() - x.min())
    f_i = lambda x: (constants.d2(x.size)**2) / (constants.d3(x.size)**2)

    # check if subgroup size is constant
    if len(set([x[~np.isnan(x)].size for x in std_obj.data])) == 1:
        # all subgroups have the same size
        # std_within = r_bar / d2(n_i)
        # std_within = (sum([subgroup_range(frame[col].dropna(axis=0)) for col in frame.columns]) / frame.shape[1])/ constants.d2(frame.shape[0])
        std_within = (sum([subgroup_range(x[~np.isnan(x)]) for x in std_obj.data]) / std_obj.nrows) / constants.d2(std_obj.ncols)
    else:
        # std_within = sum_i ( (f_i * r_i) / (d2(n_i)) ) / sum_i (f_i)
        # f_i = ( d2(n_i)**2 ) / ( d3(n_i)**2 )
        # std_within = sum([(f_i(frame[col]) * subgroup_range(frame[col])) / constants.d2(frame[col].size) for col in frame.columns])/sum([f_i(frame[col]) for col in frame.columns])
        std_within = sum([(f_i(x[~np.isnan(x)]) * subgroup_range(x[~np.isnan(x)])) / constants.d2(x.size) for x in std_obj.data]) / sum([f_i(x[~np.isnan(x)]) for x in std_obj.data])
    return std_within

def _s_bar(std_obj, **kwargs):
    """
    Calculate s_bar; Average of subgroup standard deviations for pandas DataFrame columns.
    s_bar is used for estimating std_within when subgroup size > 1
        arguments
        frame:      Pandas Dataframe
        columns:    list of colums; if not provided all columns will be treated as subgroups
    
    """
    
    # calculate overall mean
    arr = std_obj.data.flatten()
    Xi_bar = arr[~np.isnan(arr)].mean()

    # setup lambda functions
    s_i = lambda x: (sum( (x - Xi_bar)**2 ) / (x.size - 1) )**0.5
    h_i = lambda x: (constants.c4(x.size)**2) / (1 - (constants.c4(x.size)**2))
        
    # check if subgroup size is constant
    if len(set([x[~np.isnan(x)].size for x in std_obj.data])) == 1:
        # all subgroups have the same size
        # std_within = s_bar / c4(n_i)
        # std_within = (sum([frame[col].dropna(axis=0).std(ddof=0) for col in frame.columns]) / frame.shape[1]) / constants.c4(n_i)
        std_within = (sum([x[~np.isnan(x)].std(ddof=0) for x in std_obj.data]) / std_obj.nrows) / constants.c4(std_obj.ncols)
    else:
        # std_within = sum (h_i * s_i / c4(n_i)) / sum (h_i)
        # std_within = sum([(h_i(frame[col].dropna(axis=0)) * s_i(frame[col].dropna(axis=0))) / constants.c4(frame[col].dropna(axis=0).size) for col in frame.columns]) / sum(h_i(frame[col].dropna(axis=0)) for col in frame.columns)
        std_within = sum([(h_i(x[~np.isnan(x)]) * s_i(x[~np.isnan(x)])) / constants.c4(x.size) for x in std_obj.data]) / sum([h_i(x[~np.isnan(x)]) for x in std_obj.data])
    return std_within




import pandas as pd
import numpy as np


class TimeLine(pd.Series):
    def __init__(self, *args, **kwargs):
        dtype = kwargs.pop('dtype', 'float64')
        name = kwargs.pop('name', 'value')
        super().__init__(*args, name=name, dtype=dtype, **kwargs)

    def add(self, other=None, start=None, end=None, value=None, how=None, **kwargs):
        """
            Add data to the TimeLine object.
        
            Parameters
            ----------
            self: Bluebelt TimeLine object
            other: the object to add, default None
                If the object is a TimeLine object a new object is created with
                a combined index.
                If the object is a Pandas Dataframe and 'start' and 'end' are
                provided every row in the dataframe is added to the timeline. A
                'value' column is optional.
                If the object is a tuple or list and the first two items in it are
                Pandas Timestamps or can be converted to Pandas Timestamps a new
                object is created with a combined index. If a third item is given
                the value will be the third items value, if not it will be 1.
                If the object is a dict with keys 'start' and 'end' that have
                Pandas Timestamp-like values a new object is created with a combined
                index. If a 'value' key is not in the dict the value will be 1.
                If the object is a Pandas Series of int the default Pandas add
                function is used.
            start: the column name with the start Timestamp data, default None
                Only used if 'other' is a Pandas DataFrame
            end: the column name with the end Timestamp data, default None
                Only used if 'other' is a Pandas DataFrame
            value: the columns with the value data, default None
                Only used if 'other' is a Pandas DataFrame
            how: how to calculate when other is a Pandas Dataframe, default None
                If how='slow' a loop over all the values will be used which, as
                we know, one should never do. Otherwise a faster but memory-eating
                groupby method is tried. It is about three times faster but less
                memory-friendly.

            Returns
            -------
            The new Bluebelt TimeLine or Pandas Series object

        """
        if isinstance(other, TimeLine):
                index = self.index.union(other.index)

                # reindex the TimeLine objects (returns pd.Series objects)
                t1 = self.reindex(index).ffill().fillna(0)
                t2 = other.reindex(index).ffill().fillna(0)

                return TimeLine(t1.add(t2))
        elif isinstance(other, (dict, tuple, list, np.ndarray)):
            # check if there are at least two timestamps or force an error
            if isinstance(other, dict):
                start = pd.Timestamp(other.get('start'))
                end = pd.Timestamp(other.get('end'))
                value = other.get('value', 1)
            else:
                start = pd.Timestamp(other[0])
                end = pd.Timestamp(other[1])
                value = other[2] if len(other)>2 else 1
            
            other_index = pd.Index([start, end], dtype='datetime64[ns]')
            new_index = self.index.union(other_index)

            # reindex the TimeLine objects (returns pd.Series objects)
            t1 = self.reindex(new_index).ffill().fillna(0)
            t2 = pd.Series(
                index = other_index,
                data = [value, 0],
            ).reindex(new_index).ffill().fillna(0)

            return TimeLine(t1.add(t2))

        # slow but sort of memory friendly
        elif isinstance(other, pd.DataFrame) and how == 'slow' and start is not None and end is not None:
            if value:
                # sum the values
                series = other.groupby([start, end])[value].sum().rename('value').reset_index().dropna()
            else:
                # count the rows
                series = other.groupby([start, end])[start].count().rename('value').reset_index().dropna()
            
            # remove rows with start == end
            series = series[series[start]!=series[end]]

            row_count = series.shape[0]

            for i, row in enumerate(series.values):
                start = pd.Timestamp(row[0])
                end = pd.Timestamp(row[1])
                value = row[2] if len(row)>2 else 1

                other_index = pd.Index([start, end], dtype='datetime64[ns]')
                new_index = self.index.union(other_index)

                # reindex the TimeLine objects (returns pd.Series objects)
                t1 = self.reindex(new_index).ffill().fillna(0)
                t2 = pd.Series(
                    index = other_index,
                    data = [value, 0],
                ).reindex(new_index).ffill().fillna(0)

                self = TimeLine(t1.add(t2))

                if i in [int(x) for x in np.arange(0, row_count+1, row_count/100)]:
                    print(f'{i/row_count:1.0%} added', end='\r')
            
            print('100% added', end='\r')
                
            return self
        
        # a bit less slow
        elif isinstance(other, pd.DataFrame) and start is not None and end is not None:
            
            # remove rows with start == end
            # remove NaT values
            other = other[other[start]!=other[end]]
            other = other[~(other[start].isna() | other[end].isna())]
            
            # set new index
            index = pd.Index(other[start].values).union(pd.Index(other[end].values))
            
            # groupby function
            def tl(group):
                return pd.Series(index=index, data=index.to_series().between(*group.index[0], inclusive='left')) * group.values[0] 
            
            if value:
                grouped = pd.DataFrame(other.groupby([start, end])[value].sum()).groupby(level=[0,1])
            else:
                grouped = pd.DataFrame(other.groupby([start, end])[start].count()).groupby(level=[0,1])

            return self.add(TimeLine(grouped.apply(tl).T.sum(axis=1)).reduce())

        else:
            return TimeLine(super().add(other, **kwargs))
    
    def subtract(self, other=None, start=None, end=None, value=None, how=None, **kwargs):
        """
            Subtract data from the TimeLine object.
        
            Parameters
            ----------
            self: Bluebelt TimeLine object
            other: the object to subtract, default None
                If the object is a TimeLine object a new object is created with
                a combined index.
                If the object is a Pandas Dataframe and 'start' and 'end' are
                provided every row in the dataframe is added to the timeline. A
                'value' column is optional.
                If the object is a tuple or list and the first two items in it are
                Pandas Timestamps or can be converted to Pandas Timestamps a new
                object is created with a combined index. If a third item is given
                the value will be the third items value, if not it will be 1.
                If the object is a dict with keys 'start' and 'end' that have
                Pandas Timestamp-like values a new object is created with a combined
                index. If a 'value' key is not in the dict the value will be 1.
                If the object is a Pandas Series the default Pandas subtract
                function is used.
            start: the column name with the start Timestamp data, default None
                Only used if 'other' is a Pandas DataFrame
            end: the column name with the end Timestamp data, default None
                Only used if 'other' is a Pandas DataFrame
            value: the columns with the value data, default None
                Only used if 'other' is a Pandas DataFrame
            how: how to calculate when other is a Pandas Dataframe, default None
                If how='slow' a loop over all the values will be used which, as
                we know, one should never do. Otherwise a faster but memory-eating
                groupby method is tried. It is about three times faster but less
                memory-friendly.

            Returns
            -------
            The new Bluebelt TimeLine or Pandas Series object

        """
        if isinstance(other, TimeLine):
                index = self.index.union(other.index)

                # reindex the TimeLine objects (returns pd.Series objects)
                t1 = self.reindex(index).ffill().fillna(0)
                t2 = other.reindex(index).ffill().fillna(0)

                return TimeLine(t1.subtract(t2))
        elif isinstance(other, (dict, tuple, list, np.ndarray)):
            # check if there are at least two timestamps or force an error
            if isinstance(other, dict):
                start = pd.Timestamp(other.get('start'))
                end = pd.Timestamp(other.get('end'))
                value = other.get('value', 1)
            else:
                start = pd.Timestamp(other[0])
                end = pd.Timestamp(other[1])
                value = other[2] if len(other)>2 else 1
            
            other_index = pd.Index([start, end], dtype='datetime64[ns]')
            new_index = self.index.union(other_index)

            # reindex the TimeLine objects (returns pd.Series objects)
            t1 = self.reindex(new_index).ffill().fillna(0)
            t2 = pd.Series(
                index = other_index,
                data = [value, 0],
            ).reindex(new_index).ffill().fillna(0)

            return TimeLine(t1.subtract(t2))

        # slow but sort of memory friendly
        elif isinstance(other, pd.DataFrame) and how == 'slow' and start is not None and end is not None:
            if value:
                # sum the values
                series = other.groupby([start, end])[value].sum().rename('value').reset_index().dropna()
            else:
                # count the rows
                series = other.groupby([start, end])[start].count().rename('value').reset_index().dropna()
            
            # remove rows with start == end
            series = series[series[start]!=series[end]]

            row_count = series.shape[0]

            for i, row in enumerate(series.values):
                start = pd.Timestamp(row[0])
                end = pd.Timestamp(row[1])
                value = row[2] if len(row)>2 else 1

                other_index = pd.Index([start, end], dtype='datetime64[ns]')
                new_index = self.index.union(other_index)

                # reindex the TimeLine objects (returns pd.Series objects)
                t1 = self.reindex(new_index).ffill().fillna(0)
                t2 = pd.Series(
                    index = other_index,
                    data = [value, 0],
                ).reindex(new_index).ffill().fillna(0)

                self = TimeLine(t1.subtract(t2))

                if i in [int(x) for x in np.arange(0, row_count+1, row_count/100)]:
                    print(f'{i/row_count:1.0%} subtracted', end='\r')
            
            print('100% subtracted', end='\r')
                
            return self
        
        # a bit less slow
        elif isinstance(other, pd.DataFrame) and start is not None and end is not None:
            
            # remove rows with start == end
            # remove NaT values
            other = other[other[start]!=other[end]]
            other = other[~(other[start].isna() | other[end].isna())]
            
            # set new index
            index = pd.Index(other[start].values).union(pd.Index(other[end].values))
            
            # groupby function
            def tl(group):
                return pd.Series(index=index, data=index.to_series().between(*group.index[0], inclusive='left')) * group.values[0] 
            
            if value:
                grouped = pd.DataFrame(other.groupby([start, end])[value].sum()).groupby(level=[0,1])
            else:
                grouped = pd.DataFrame(other.groupby([start, end])[start].count()).groupby(level=[0,1])

            return self.subtract(TimeLine(grouped.apply(tl).T.sum(axis=1)).reduce())
            
        else:
            return super().subtract(other, **kwargs)
        
    def reduce(self, *args, **kwargs):
        """
            Reduce the TimeLine object by grouping the index by value.
        
            Parameters
            ----------
            self: Bluebelt TimeLine object

            Returns
            -------
            The Bluebelt TimeLine object with a new index

        """
        return TimeLine(
            index = self.index[(self.values!=self.shift().values)],
            data = self.groupby((self.values!=self.shift().values).cumsum()).max().values            
        )
    
    def resample(self, rule='15T', *args, **kwargs):
        """
            Resample the TimeLine object.
        
            Parameters
            ----------
            self: Bluebelt TimeLine object
            rule: Pandas resample rule, default '15T'

            Returns
            -------
            The Bluebelt TimeLine object with a new index

        """
        # get the gcd in nanoseconds
        gcd = np.gcd.reduce(
            [int(x / np.timedelta64(1, 'ns')) for x in (self.index[1:] - self.index[:-1]).values]
        )
        
        # check if the rule >= gcd
        if pd.Timedelta(rule) >= np.timedelta64(gcd, 'ns'):
            gcd_rule = f'{gcd}ns'
        else: # rule < gcd
            gcd_rule = rule
            
        # first resample to gcd with ffill, then to desired rule with max
        return TimeLine(super().resample(gcd_rule).ffill().resample(rule).max())
    
    def reshape(self, rule='week', level='minute', aggfunc='max'):
        """
            Reshape the TimeLine object.
        
            Parameters
            ----------
            self: Bluebelt TimeLine object
            rule: reshaping rule, default 'week'
                Indicates the new index level, 'year', 'week', 'day', 'hour',
                'minute' or 'second'. The lower levels are converted to columns.
            level: the lowest level to return, default 'minute'
            aggfunc: the Pandas.pivot_table aggfunc, default 'max'
            
            Returns
            -------
            a Pandas DataFrame object

        """
        name = self.name
        levels = ['year', 'week', 'day', 'hour', 'minute', 'second']
        
        if not (rule in levels and level in levels):
            raise ValueError("Both rule and level should be 'year', 'week', 'day', 'hour', 'minute' or 'second'")
            
        # avoid trouble if the level has a lower index than rule 
        level = level if levels.index(rule) < levels.index(level) else rule
        
        frame = pd.DataFrame(self)
        frame.loc[:,['year', 'week', 'day']] = frame.index.isocalendar()
        frame['hour'] = frame.index.hour
        frame['minute'] = frame.index.minute
        frame['second'] = frame.index.second
        return frame.pivot_table(
            index=levels[:levels.index(rule)+1],
            columns=levels[levels.index(rule)+1:levels.index(level)+1],
            values=name,
            aggfunc=aggfunc,
        )
    
    def percentile(self, percentile=0.1, how='smallest', rule='week', level='minute'):
        """
            Reshape the TimeLine object and return a pandas Series for one period.
            The period is constructed as the percentile for every 'level' time slot
            over one 'rule' period.
        
            Parameters
            ----------
            self: Bluebelt TimeLine object
            percentile: the percentile to calculate, default 0.1
            how: calculate the percentile as 'smallest' or 'largest', default 'smallest'
            rule: reshaping rule, default 'week'
                Indicates the new index level, 'year', 'week', 'day', 'hour',
                'minute' or 'second'. The lower levels are converted to columns.
            level: the lowest level to return, default 'minute'
            
            Returns
            -------
            a Pandas DataFrame object

        """
        levels = ['year', 'week', 'day', 'hour', 'minute', 'second']
        
        # avoid trouble if the level has a lower index than rule 
        level = level if levels.index(rule) < levels.index(level) else levels[levels.index(rule)+1]
        
        # yes, reshape to the lowest level and return a series
        result = self.reshape(rule=level).iloc[:,0]
        
        # get n from percentile
        n = np.round(result.groupby(level=levels[:levels.index(rule)+1]).ngroups * percentile).astype(int)
        
        # set groupby_levels
        groupby_levels = levels[levels.index(rule)+1:levels.index(level)+1]
        
        # and now groupby levels down to 'rule'
        if how == 'largest':
            result = result.groupby(level=groupby_levels, group_keys=False).nlargest(n).groupby(level=groupby_levels, group_keys=False).nsmallest(1)
        elif how == 'smallest':
            result = result.groupby(level=groupby_levels, group_keys=False).nsmallest(n).groupby(level=groupby_levels, group_keys=False).nlargest(1)
        else:
            raise ValueError(f"parameter 'how' must be 'smallest' or 'largest', not {how}.")
            
        return result


import pandas as pd
import bluebelt.analysis.performance

def performance(cls):
    def summary(self, values=None, **kwargs):
        check_attr(self, values)
        return bluebelt.analysis.performance.Summary(eval(f"self.{values}"), **kwargs)        

    def control_chart(self, values=None, **kwargs):
        check_attr(self, values)
        return bluebelt.analysis.performance.ControlChart(eval(f"self.{values}"), **kwargs)
        
    def run_chart(self, values=None, **kwargs):
        check_attr(self, values)
        return bluebelt.analysis.performance.RunChart(eval(f"self.{values}"), **kwargs)
    
    def process_capability(self, values=None, **kwargs):
        check_attr(self, values)
        return bluebelt.analysis.performance.ProcessCapability(eval(f"self.{values}"), **kwargs)
    
    def check_attr(self, values):
        attributes = []
        for attr, value in self.__dict__.items():
            if isinstance(value, pd.Series):
                attributes += ["'"+attr+"'"]
        attributes = ", ".join(attributes)
            
        if not values:
            raise ValueError(f"you must provide a 'values' parameter, please choose from {attributes}")
        if not hasattr(self, values):
           
            if len(attributes)>0:
                raise ValueError(f"'{values}' is not a valid attribute, please choose from {attributes}")
            else:
                raise ValueError(f"there are no valid attributes available, the performance function will not work...")
        
    setattr(cls, "summary", summary)
    setattr(cls, "control_chart", control_chart)
    setattr(cls, "run_chart", run_chart)
    setattr(cls, "process_capability", process_capability)
    
    return cls



import os
import pandas as pd
import numpy as np
import scipy.stats as stats
import datetime
import warnings

from matplotlib.colors import to_rgba
import matplotlib.pyplot as plt
import matplotlib.dates
import plotly.graph_objects as go
import bluebelt.helpers.matplotlib2plotly as m2p


from bluebelt.helpers.date import year_with_most_iso_weeks
import bluebelt.helpers.ticks

from bluebelt.helpers.decorators.performance import performance
from bluebelt.helpers.decorators.pattern import pattern


@pattern
@performance
class Polynomial:
    def __init__(
        self,
        _obj,
        shape=(0, 6),
        validation="rsq",
        threshold=0.05,
        confidence=0.8,
        outlier_sigma=2,
        adjust=True,
        **kwargs,
    ):

        self._obj = _obj
        self.name = _obj.name or "obj"
        self.shape = shape
        self.validation = validation
        self.threshold = threshold
        self.confidence = confidence
        self.outlier_sigma = outlier_sigma
        self.adjust = adjust

        self.calculate()

    def calculate(self):

        # set pattern and residuals
        _poly_hand_granade(self)

        # set outliers
        _set_outliers(self)

        # handle adjusted
        self.adjusted = self._obj.loc[~self.outliers.notnull()]
        if self.adjust:

            self.__obj = self._obj.copy()  # backup

            # replace outliers with None values so they will be ignored by _poly_hand_granade and reset pattern
            self._obj = pd.Series(
                data=np.where(self.outliers.notnull(), None, self._obj).astype(
                    np.float
                ),
                index=self._obj.index,
            )
            _poly_hand_granade(self)

            self._obj = self.__obj.copy()  # and reset
            del self.__obj

        # handle bounds
        _set_bounds(self)

        # set final shape
        if hasattr(self, "_shape"):
            self.shape = self._shape
            del self._shape

    def set_observations(self):
        _set_observations(self)

    def set_residuals(self):
        _set_residuals(self)

    def set_outliers(self):
        _set_outliers(self)

    def set_bounds(self):
        _set_bounds(self)

    def __repr__(self):
        return f"{self.__class__.__name__}(n={self._obj.size:1.0f}, shape={self.shape}, validation='{self.validation}', threshold={self.threshold}, confidence={self.confidence}, outlier_sigma={self.outlier_sigma}, adjust={self.adjust}, outliers={self.outliers_count}, rsq={self.rsq:1.2f}, std={self.std:1.2f}, p_value={self.p_value:1.2f})"

    def __str__(self):
        _result = f"-" * (len(self.name) + 4) + "\n"
        _result += f"  {self.name}\n"
        _result += f"-" * (len(self.name) + 4) + "\n"
        _result += f"\n"

        _result += f"input variables\n"
        _result += f"-" * 50 + "\n"
        _result += f'  {"_obj size:":<30}{self._obj.size:1.0f}\n'
        _result += f'  {"validation type:":<30}{self.validation}\n'
        _result += f'  {"validation threshold:":<30}{self.threshold:1.4f}\n'

        _result += f"\n"
        _result += f"pattern\n"
        _result += f"-" * 50 + "\n"
        _result += f'  {"shape:":<30}{self.shape:1.0f}\n'
        _result += f'  {"r squared:":<30}{self.rsq:1.2f}\n'

        _result += f"\n"
        _result += f"residuals\n"
        _result += f"-" * 50 + "\n"
        _result += f'  {"bounds level:":<30}{self.confidence * 100:1.0f}%\n'
        _result += f'  {"bounds size (upper-lower):":<30}{self.bounds * 2:1.2f}\n'
        _result += f'  {"standard deviation:":<30}{self.std:1.2f}\n'
        _result += f'  {"p-value normal distribution:":<30}{self.p_value:1.4f}\n'
        _result += f'  {"outliers:":<30}{self.outliers_count:1.0f}\n'

        return _result

    def plot(
        self,
        bounds=True,
        residuals=False,
        xlim=(None, None),
        ylim=(None, None),
        max_xticks=None,
        format_xticks=None,
        format_yticks=None,
        title=None,
        xlabel=None,
        ylabel=None,
        legend=True,
        path=None,
        **kwargs,
    ):
        return _plot(
            self,
            bounds=bounds,
            residuals=residuals,
            xlim=xlim,
            ylim=ylim,
            max_xticks=max_xticks,
            format_xticks=format_xticks,
            format_yticks=format_yticks,
            title=title,
            xlabel=xlabel,
            ylabel=ylabel,
            legend=legend,
            path=path,
            **kwargs,
        )

@pattern
@performance
class Periodical:
    def __init__(
        self,
        _obj,
        rule="1W",
        how="mean",
        confidence=0.8,
        outlier_sigma=2,
        adjust=True,
        **kwargs,
    ):

        self._obj = _obj
        self.name = _obj.name or "obj"
        self.rule = rule
        self.how = how
        self.confidence = confidence
        self.outlier_sigma = outlier_sigma
        self.adjust = adjust

        self.calculate()

    def calculate(self):

        # set pattern and residuals
        _peri_hand_granade(self)

        # set outliers
        self.outliers = pd.Series(
            data=np.where(
                self.residuals.abs() > self.residuals.std() * self.outlier_sigma,
                self._obj,
                None,
            ),
            index=self._obj.index,
            name=f"{self._obj.name} {self.rule} outliers",
        )
        self.outliers_count = np.count_nonzero(self.outliers)

        # handle adjusted
        self.adjusted = self._obj.loc[~self.outliers.notnull()]
        if self.adjust:

            self.__obj = self._obj.copy()  # backup

            # replace outliers with None values so they will be ignored by _poly_hand_granade and reset pattern
            self._obj = pd.Series(
                data=np.where(self.outliers.notnull(), None, self._obj).astype(
                    np.float
                ),
                index=self._obj.index,
                name=self._obj.name,
            )
            _peri_hand_granade(self)

            self._obj = self.__obj.copy()  # and reset
            del self.__obj

        # handle bounds
        _set_bounds(self)

    def set_observations(self):
        _set_observations(self)

    def set_residuals(self):
        _set_residuals(self)

    def set_outliers(self):
        _set_outliers(self)

    def set_bounds(self):
        _set_bounds(self)

    def __repr__(self):
        return f"{self.__class__.__name__}(n={self._obj.size:1.0f}, rule={self.rule}, how={self.how}, confidence={self.confidence}, outlier_sigma={self.outlier_sigma}, adjust={self.adjust}, outliers={self.outliers_count})"

    def plot(
        self,
        bounds=True,
        residuals=False,
        xlim=(None, None),
        ylim=(None, None),
        max_xticks=None,
        format_xticks=None,
        format_yticks=None,
        title=None,
        xlabel=None,
        ylabel=None,
        legend=True,
        path=None,
        **kwargs,
    ):
        return _plot(
            self,
            bounds=bounds,
            residuals=residuals,
            xlim=xlim,
            ylim=ylim,
            max_xticks=max_xticks,
            format_xticks=format_xticks,
            format_yticks=format_yticks,
            title=title,
            xlabel=xlabel,
            ylabel=ylabel,
            legend=legend,
            path=path,
            **kwargs,
        )

def _poly_hand_granade(_poly):

    # get index and _index
    index = matplotlib.dates.date2num(
        _poly._obj.index
    )  # bluebelt.core.index.IndexToolkit(_poly._obj.index).clean()
    _index = matplotlib.dates.date2num(
        _poly._obj.dropna().index
    )  # bluebelt.core.index.IndexToolkit(_poly._obj.dropna().index).clean()

    # get the values
    _values = _poly._obj.dropna().values

    if isinstance(_poly.shape, int):
        polyfit = np.polynomial.polynomial.polyfit(_index, _values, _poly.shape)
        _poly.pattern = pd.Series(
            index=_poly._obj.index,
            data=np.polynomial.polynomial.polyval(index, polyfit),
            name=f"{_poly.name} {_get_nice_polynomial_name(_poly.shape)}",
        )
        _poly.residuals = (_poly._obj - _poly.pattern).rename(
            f"{_poly.name} {_get_nice_polynomial_name(_poly.shape)} residuals"
        )

        _poly.statistic, _poly.p_value = stats.normaltest(
            _poly.residuals.dropna().values
        )
        np_err = np.seterr(
            divide="ignore", invalid="ignore"
        )  # ignore possible divide by zero
        _poly.rsq = (
            np.corrcoef(
                _poly._obj.dropna().values, _poly.pattern[~_poly._obj.isna()].values
            )[1, 0]
            ** 2
        )
        np.seterr(**np_err)  # go back to previous settings
        _poly.std = _poly.residuals.std()

    elif isinstance(_poly.shape, tuple):

        _results = {}
        _validation = {}

        with warnings.catch_warnings():
            warnings.filterwarnings("error")
            for shape in range(_poly.shape[0], _poly.shape[1] + 1):
                try:
                    polyfit = np.polynomial.polynomial.polyfit(_index, _values, shape)
                    pattern = pd.Series(
                        index=_poly._obj.index,
                        data=np.polynomial.polynomial.polyval(index, polyfit),
                        name=f"{_poly.name} {_get_nice_polynomial_name(shape)}",
                    )
                    residuals = (_poly._obj - pattern).rename(
                        f"{_poly.name} {_get_nice_polynomial_name(shape)} residuals"
                    )

                    statistic, p_value = stats.normaltest(residuals.dropna().values)
                    np_err = np.seterr(
                        divide="ignore", invalid="ignore"
                    )  # ignore possible divide by zero
                    rsq = (
                        np.corrcoef(
                            _poly._obj.dropna().values,
                            pattern[~_poly._obj.isna()].values,
                        )[1, 0]
                        ** 2
                    )
                    np.seterr(**np_err)  # go back to previous settings
                    std = residuals.std()

                    _results[shape] = {
                        "pattern": pattern,
                        "residuals": residuals,
                        "statistic": statistic,
                    }
                    _validation[shape] = {
                        "p_value": p_value,
                        "rsq": rsq,
                        "std": std,
                    }

                except np.polynomial.polynomial.pu.RankWarning:
                    # print(f'RankWarning at {_get_nice_polynomial_name(shape)}')
                    pass  # break

        if _poly.validation == "p_value":
            validation = pd.DataFrame.from_dict(_validation).loc[_poly.validation]

            # if any p_value >= threshold then we have a winner
            if (validation >= _poly.threshold).any():
                shape = validation.idxmax()
            else:
                shape = 0

        elif _poly.validation == "std":

            # we want a small std of the residuals
            validation = pd.DataFrame.from_dict(_validation).loc[_poly.validation]

            # are there any relevant improvements?
            if (validation.diff().abs() / validation.shift() >= _poly.threshold).any():
                # find the relevant improvements
                relevant = validation.iloc[validation.diff().idxmin() :]

                # does it get any better than this?
                improvements = (
                    relevant.diff() / relevant.shift(1)
                ).abs() >= _poly.threshold
                improvements.at[validation.diff().idxmin()] = True  # because it must be

                # where does it stop getting better?
                first_fail = (
                    improvements.where(improvements == False).first_valid_index()
                    or improvements.index.max() + 1
                )

                # so what is the shape?
                shape = improvements[improvements.index < first_fail].index.max()
            else:
                shape = validation.index[0]

        else:  # the default
            _poly.validation = "rsq"

            # we want a big rsq value
            validation = pd.DataFrame.from_dict(_validation).loc[_poly.validation]

            # are there any relevant improvements?
            if (validation.diff().abs() / validation.shift() >= _poly.threshold).any():
                # find the relevant improvements
                relevant = validation.iloc[validation.diff().idxmax() :]

                # does it get any better than this?
                improvements = (
                    relevant.diff() / relevant.shift(1)
                ).abs() >= _poly.threshold
                improvements.at[validation.diff().idxmax()] = True  # because it must be

                # where does it stop getting better?
                first_fail = (
                    improvements.where(improvements == False).first_valid_index()
                    or improvements.index.max() + 1
                )

                # so what is the shape?
                shape = improvements[improvements.index < first_fail].index.max()
            else:
                shape = validation.index[0]

        _poly.pattern = _results.get(shape).get("pattern")
        _poly.residuals = _results.get(shape).get("residuals")
        _poly._shape = shape
        _poly.statistic = _results.get(shape).get("statistic")
        _poly.p_value = _validation.get(shape).get("p_value")
        _poly.rsq = _validation.get(shape).get("rsq")
        _poly.std = _validation.get(shape).get("std")

    else:
        _poly.pattern = None
        _poly.residuals = None

    return

def _peri_hand_granade(_peri, **kwargs):

    # set pattern and residuals
    if _peri.how == "mean":
        _peri.pattern = bluebelt.data.resolution.Flatten(
            _peri._obj, _peri.rule, **kwargs
        ).mean()
    elif _peri.how == "min":
        _peri.pattern = bluebelt.data.resolution.Flatten(
            _peri._obj, _peri.rule, **kwargs
        ).min()
    elif _peri.how == "max":
        _peri.pattern = bluebelt.data.resolution.Flatten(
            _peri._obj, _peri.rule, **kwargs
        ).max()
    elif _peri.how == "std":
        _peri.pattern = bluebelt.data.resolution.Flatten(
            _peri._obj, _peri.rule, **kwargs
        ).std()
    else:
        _peri.pattern = bluebelt.data.resolution.Flatten(
            _peri._obj, _peri.rule, **kwargs
        ).sum()

    _peri.residuals = _peri._obj - _peri.pattern

    _peri.statistic, _peri.p_value = stats.normaltest(_peri.residuals.dropna().values)
    _peri.rsq = np.corrcoef(_peri._obj.values, _peri.pattern.values)[1, 0] ** 2

    return

def _set_observations(_poly):
    _poly._obj = (_poly.pattern + _poly.residuals).rename(_poly._obj.name)

def _set_residuals(_poly):
    _poly.residuals = (_poly._obj - _poly.pattern).rename(f"residuals")

    _poly.statistic, _poly.p_value = stats.normaltest(_poly.residuals.dropna().values)
    np_err = np.seterr(
        divide="ignore", invalid="ignore"
    )  # ignore possible divide by zero
    _poly.rsq = (
        np.corrcoef(
            _poly._obj.dropna().values,
            _poly.pattern[~_poly._obj.isna()].values,
        )[1, 0]
        ** 2
    )
    np.seterr(**np_err)  # go back to previous settings
    _poly.std = _poly.residuals.std()

def _set_outliers(_poly):
    # set outliers
    _poly.outliers = pd.Series(
        data=np.where(
            _poly.residuals.abs() > _poly.residuals.std() * _poly.outlier_sigma,
            _poly._obj,
            None,
        ),
        index=_poly._obj.index,
    )
    _poly.outliers_count = np.count_nonzero(_poly.outliers)

    return _poly

def _set_bounds(_poly):

    _poly.sigma_level = stats.norm.ppf(1 - (1 - _poly.confidence) / 2)

    # set bounds
    _poly.upper = _poly.pattern + _poly.residuals.std() * _poly.sigma_level
    _poly.lower = _poly.pattern - _poly.residuals.std() * _poly.sigma_level
    _poly.bounds = _poly.residuals.std() * _poly.sigma_level

    # set out of bounds values
    _poly.out_of_bounds = _poly._obj.where(
        ((_poly._obj > _poly.upper) | (_poly._obj < _poly.lower))
        & (_poly.outliers.isnull()),
        np.nan,
    )
    _poly.within_bounds = _poly._obj.where(
        (_poly._obj <= _poly.upper) & (_poly._obj >= _poly.lower), np.nan
    )

    return _poly

def _get_nice_polynomial_name(shape):
    if shape == -1:
        return "unknown"
    if shape == 0:
        return "linear"
    if shape == 1:
        return str(shape) + "st degree polynomial"
    elif shape == 2:
        return str(shape) + "nd degree polynomial"
    elif shape == 3:
        return str(shape) + "rd degree polynomial"
    else:
        return str(shape) + "th degree polynomial"

def _plot(
    _poly,
    bounds=True,
    residuals=False,
    xlim=(None, None),
    ylim=(None, None),
    max_xticks=None,
    format_xticks=None,
    format_yticks=None,
    title=None,
    xlabel=None,
    ylabel=None,
    legend=True,
    path=None,
    **kwargs,
):

    title = title if title is not None else  f"{_poly.pattern.name}"
    bins = kwargs.get('bins', None)

    if bluebelt.config('plotting') == 'matplotlib':
        # prepare figure
        fig, ax = plt.subplots()
        
        if residuals:
            # residuals histogram
            ax.hist(_poly.residuals.values, bins=bins, density=True, **bluebelt.style("pattern.residuals.hist"))
            ax.set_yticks([])

            # get current limits
            xlims = ax.get_xlim()

            # fit a normal distribution to the data
            norm_mu, norm_std = stats.norm.fit(_poly.residuals.dropna())
            pdf_x = np.linspace(xlims[0], xlims[1], 100)
            ax.plot(
                pdf_x,
                stats.norm.pdf(pdf_x, norm_mu, norm_std),
                **bluebelt.style("pattern.residuals.normal"),
            )

            # histogram x label
            ax.set_xlabel("residuals distribution")

            ax.text(
                0.02,
                0.98,
                f"D'Agostino-Pearson\nstatistic: {_poly.statistic:1.2f}\np: {_poly.p_value:1.2f}",
                transform=ax.transAxes,
                **bluebelt.style("pattern.statistics"),
            )
        else:
            
            # pattern
            ax.plot(
                _poly.pattern.index,
                _poly.pattern.values,
                label="pattern",
                **bluebelt.style("pattern.line"),
            )

            if bounds:
                # observations
                ax.plot(
                    _poly.pattern.index,
                    _poly.within_bounds.values,
                    label="observations",
                    **bluebelt.style("pattern.observations"),
                )

                # bounds (fill between and lines)
                ax.fill_between(
                    _poly.pattern.index,
                    _poly.lower.values,
                    _poly.upper.values,
                    label=f"{(_poly.confidence * 100):1.0f}% bounds",
                    **bluebelt.style("pattern.bounds.fill"),
                )
                ax.plot(
                    _poly.pattern.index,
                    _poly.lower.values,
                    label=None,
                    **bluebelt.style("pattern.bounds.line"),
                )
                ax.plot(
                    _poly.pattern.index,
                    _poly.upper.values,
                    label=None,
                    **bluebelt.style("pattern.bounds.line"),
                )

                # out of bounds
                ax.plot(
                    _poly.pattern.index,
                    _poly.out_of_bounds.values,
                    label="out of bounds",
                    **bluebelt.style("pattern.out_of_bounds"),
                )
            else:
                # observations
                ax.plot(
                    _poly.pattern.index,
                    _poly.within_bounds.values,
                    label="observations",
                    **bluebelt.style("pattern.observations"),
                )

                # out of bounds
                ax.plot(
                    _poly.pattern.index,
                    _poly.out_of_bounds.values,
                    label="out of bounds",
                    **bluebelt.style("pattern.observations"),
                )

            # outliers
            ax.plot(
                _poly.pattern.index,
                _poly.outliers.values,
                label="outliers",
                **bluebelt.style("pattern.outliers"),
            )

            # set ticks
            if max_xticks is None:
                max_xticks = bluebelt.helpers.ticks.get_max_xticks(ax)
            bluebelt.helpers.ticks.year_week(_poly.pattern, ax=ax, max_xticks=max_xticks)

            # format ticks
            if format_xticks:
                ax.set_xticks(ax.get_xticks())
                ax.set_xticklabels([f"{x:{format_xticks}}" for x in ax.get_xticks()])
            if format_yticks:
                ax.set_yticks(ax.get_yticks())
                ax.set_yticklabels([f"{y:{format_yticks}}" for y in ax.get_yticks()])

            # legend
            if legend:
                ax.legend()
            elif ax.get_legend() is not None:
                ax.get_legend().set_visible(False)

        # limit axis
        ax.set_xlim(xlim)
        ax.set_ylim(ylim)

        # labels
        if title:
            ax.set_title(title)
        if xlabel:
            ax.set_xlabel(xlabel)
        if ylabel:
            ax.set_ylabel(ylabel)


        plt.tight_layout()

        # file
        if path:
            if len(os.path.dirname(path)) > 0 and not os.path.exists(os.path.dirname(path)):
                os.makedirs(os.path.dirname(path))
            plt.savefig(path)
            plt.close()
        else:
            plt.close()
            return fig

    elif bluebelt.config('plotting') == 'plotly':

        layout = go.Layout(
                m2p.layout(
                    title=title,
                    xlabel=xlabel,
                    ylabel=ylabel,
                    xlim=xlim,
                    ylim=ylim,
                    format_xticks=format_xticks,
                    format_yticks=format_yticks,
                    legend=legend,
                    **kwargs,
                )
            )
        if residuals:
            # normal dist
            norm_mu, norm_std = stats.norm.fit(_poly.residuals.dropna())
            pdf_x = np.linspace(_poly.residuals.min(), _poly.residuals.max(), 100)

            data = [
                dict(
                    type="histogram",
                    histnorm="probability density",
                    nbinsx = bins,
                    x=_poly.residuals.values,
                    marker=dict(color=bluebelt.style("pattern.residuals.hist.facecolor"), opacity=bluebelt.style("pattern.residuals.hist.alpha")),
                    name='residuals',
                    opacity=1,
                    showlegend=False,
                ),
                dict(
                    type="scatter",
                    mode="lines",
                    x=pdf_x,
                    y=stats.norm.pdf(pdf_x, norm_mu, norm_std),
                    name="normal fit",
                    line = dict(color=bluebelt.style("pattern.residuals.normal.color"), width=bluebelt.style("pattern.residuals.normal.linewidth"), dash=m2p.linestyle(bluebelt.style("pattern.residuals.normal.linestyle"))),
                    hoverinfo="none",
                    showlegend=True,
                ),
            ]

            layout['yaxis'] = dict(ticks="", showticklabels=False)
            layout['xaxis'] = dict(showgrid=False, ticks='outside', title_text=xlabel)
            layout['bargap'] = 1 - bluebelt.style("pattern.residuals.hist.rwidth")

        else:
            data = [
                dict(
                    type="scatter",
                    mode="markers",
                    x=_poly.outliers.index.values,
                    y=_poly.outliers.values,
                    name='outliers',
                    marker=dict(size=bluebelt.style("pattern.outliers.markersize"), color=bluebelt.style("pattern.outliers.markerfacecolor")),
                    opacity=1,
                    showlegend=True,
                ),
                dict(
                    type="scatter",
                    mode="markers",
                    x=_poly.out_of_bounds.index.values,
                    y=_poly.out_of_bounds.values,
                    name='out of bounds',
                    marker=dict(size=bluebelt.style("pattern.out_of_bounds.markersize"), color=bluebelt.style("pattern.out_of_bounds.markerfacecolor")),
                    opacity=1,
                    showlegend=True,
                ),
                dict(
                    type="scatter",
                    mode="markers",
                    x=_poly.within_bounds.index.values,
                    y=_poly.within_bounds.values,
                    name='within bounds',
                    marker=dict(size=bluebelt.style("pattern.observations.markersize"), color=bluebelt.style("pattern.observations.markerfacecolor")),
                    opacity=1,
                    showlegend=True,
                ),
                dict(
                    type="scatter",
                    mode="lines",
                    x=_poly.pattern.index.values,
                    y=_poly.pattern.values,
                    name="pattern",
                    line = dict(color=bluebelt.style("pattern.line.color"), width=bluebelt.style("pattern.line.linewidth")),
                    hoverinfo="none",
                    showlegend=True,
                ),
                dict(
                    type="scatter",
                    mode="lines",
                    x=_poly.lower.index.values,
                    y=_poly.lower.values,
                    name="lower",
                    line = dict(color=bluebelt.style("pattern.bounds.line.color"), width=bluebelt.style("pattern.bounds.line.linewidth")),
                    hoverinfo="none",
                    showlegend=False,
                ),
                dict(
                    type="scatter",
                    mode="lines",
                    x=_poly.upper.index.values,
                    y=_poly.upper.values,
                    name=f"{_poly.confidence * 100:1.0f}% bounds",
                    line = dict(color=bluebelt.style("pattern.bounds.line.color"), width=bluebelt.style("pattern.bounds.line.linewidth")),
                    hoverinfo="none",
                    showlegend=True,
                    fill='tonexty',
                    fillcolor='rgba' + str(
                        to_rgba(
                            bluebelt.style("pattern.bounds.fill.color") + 
                            format(int(255*bluebelt.style("pattern.bounds.fill.alpha")), 'x')
                            )
                    ),
                ),
            ]

        fig = go.Figure(data=data, layout=layout)
        if xlim:
            fig.update_layout(xaxis_range=xlim)
        if ylim:
            fig.update_layout(yaxis_range=ylim)
        if format_xticks:
            fig.update_layout(xaxis_tickformat = format_xticks)
        if format_yticks:
            fig.update_layout(yaxis_tickformat = format_yticks)
        

        return fig
    
    else:
        return     


import numpy as np
import pandas as pd
import bluebelt.analysis.datetime
import bluebelt.data.projection
from bluebelt.helpers.decorators.docstring import docstring


def pattern(cls):
    @docstring(bluebelt.data.projection.project)
    def project(self, year=None, adjust_holidays=True):
        self._obj = bluebelt.data.projection.project(
            self._obj, year=year, adjust_holidays=adjust_holidays
        )
        self.pattern = bluebelt.data.projection.project(
            self.pattern, year=year, adjust_holidays=adjust_holidays
        )
        self.residuals = bluebelt.data.projection.project(
            self.residuals, year=year, adjust_holidays=adjust_holidays
        )
        self.outliers = bluebelt.data.projection.project(
            self.outliers, year=year, adjust_holidays=adjust_holidays
        )
        self.adjusted = bluebelt.data.projection.project(
            self.adjusted, year=year, adjust_holidays=adjust_holidays
        )
        self.upper = bluebelt.data.projection.project(
            self.upper, year=year, adjust_holidays=adjust_holidays
        )
        self.lower = bluebelt.data.projection.project(
            self.lower, year=year, adjust_holidays=adjust_holidays
        )
        self.out_of_bounds = bluebelt.data.projection.project(
            self.out_of_bounds, year=year, adjust_holidays=adjust_holidays
        )
        self.within_bounds = bluebelt.data.projection.project(
            self.within_bounds, year=year, adjust_holidays=adjust_holidays
        )

        return self

    def set_pattern(self, pattern=None):
        """
            Set the pattern of a Bluebelt pattern object (polynomial or
            periodical)
        
            Parameters
            ----------
            pattern: pandas Series, default None
                the pandas Series with the new pattern
                provide a matching DatetimeIndex

            Returns
            -------
            The Bluebelt pattern object with a new pattern

        """
        # quick fix for patterns that have more datapoints than self
        pattern = pattern[self._obj.index]

        if pattern.shape[0] < self._obj.shape[0]:
            raise ValueError(
                "The provided pattern does not match the original pattern. Please check the pattern index."
            )

        self.pattern = pattern
        self.shape = -1

        # replace the original observations
        self.set_observations()

        self.set_residuals()
        self.set_outliers()
        self.set_bounds()

        return self

    def add_weekday(self, values=None):
        values = values or self._obj
        weekday = bluebelt.analysis.datetime.WeekDay(values)

        # check if weekdays are relevamt
        if weekday.equal_means:
            raise ValueError("The weekdays have equal means.")

        self.pattern = (
            self.pattern * self.pattern.index.weekday.map(weekday.ratio).values
        )

        self.set_residuals()
        self.set_outliers()
        self.set_bounds()

        return self

    def add_monthday(self, values=None):
        values = values or self._obj
        monthday = bluebelt.analysis.datetime.MonthDay(values)

        # check if weekdays are relevamt
        if monthday.equal_means:
            raise ValueError("The monthdays have equal means.")

        self.pattern = (
            self.pattern * self.pattern.index.weekday.map(monthday.ratio).values
        )

        self.set_residuals()
        self.set_outliers()
        self.set_bounds()

        return self

    setattr(cls, "project", project)
    setattr(cls, "set_pattern", set_pattern)
    setattr(cls, "add_weekday", add_weekday)
    setattr(cls, "add_monthday", add_monthday)

    return cls



import math
import pandas as pd
import numpy as np

import matplotlib.pyplot as plt

from bluebelt.helpers.ci import ci_mean
from bluebelt.helpers.decorators.performance import performance

import bluebelt.styles

import warnings
import os

import warnings

@pd.api.extensions.register_dataframe_accessor("f")
class Forecast:
    def __init__(self, pandas_obj, **kwargs):
        self._obj = pandas_obj
        self.MAE = self.MAE(self._obj, **kwargs)
        self.MAPE = self.MAPE(self._obj, **kwargs)
        self.SMAPE = self.SMAPE(self._obj, **kwargs)
        self.MDA = self.MDA(self._obj, **kwargs)
        self.MPE = self.MPE(self._obj, **kwargs)
        self.NFM = self.NFM(self._obj, **kwargs)
        
    def __repr__(self):
        
        return f"MAE={self.MAE.result:1.2f}, MAPE={self.MAPE.result:1.2f}, SMAPE={self.SMAPE.result:1.2f}, MDA={self.MDA.result:1.2f}, MPE={self.MPE.result:1.2f}, NFM={self.NFM.result:1.2f}"

    @performance
    class MAE:
        def __init__(self, frame, forecast=None, actuals=None, **kwargs):

            self.frame = frame
            self.forecast = forecast
            self.actuals = actuals
            self.calculate()

        def calculate(self):

            _set_forecast_and_actuals(self)

            if self.forecast is not None and self.actuals is not None:
                self.result = np.mean(np.abs(self.actuals - self.forecast))
                self.values = np.abs(self.actuals - self.forecast)
            else:
                self.result = None
                self.values = None

        def __repr__(self):
            return f"{self.__class__.__name__}(n={self.frame.shape[0]:1.0f}, result={self.result:1.4f})"

        def plot(
            self,
            xlim=(None, None),
            ylim=(None, None),
            bins=20,
            format_xticks="1.0f",
            title=None,
            xlabel=None,
            ylabel=None,
            path=None,
            **kwargs,
        ):
            return _plot(
                self,
                xlim=xlim,
                ylim=ylim,
                bins=bins,
                format_xticks=format_xticks,
                title=title,
                xlabel=xlabel,
                ylabel=ylabel,
                path=path,
                **kwargs,
            )

    @performance
    class MAPE:
        def __init__(self, frame, forecast=None, actuals=None, **kwargs):

            self.frame = frame
            self.forecast = forecast
            self.actuals = actuals
            self.calculate()

        def calculate(self):

            _set_forecast_and_actuals(self)

            if self.forecast is not None and self.actuals is not None:
                self.result = (
                    np.abs((self.actuals - self.forecast) / self.actuals).sum()
                ) / len(self.forecast)
                self.values = np.abs((self.actuals - self.forecast) / self.actuals)
            else:
                self.result = None
                self.values = None

        def __repr__(self):
            return f"{self.__class__.__name__}(n={self.frame.shape[0]:1.0f}, result={self.result:1.4f})"

        def plot(
            self,
            xlim=(None, None),
            ylim=(None, None),
            bins=20,
            format_xticks="1.0%",
            title=None,
            xlabel=None,
            ylabel=None,
            path=None,
            **kwargs,
        ):
            return _plot(
                self,
                xlim=xlim,
                ylim=ylim,
                bins=bins,
                format_xticks=format_xticks,
                title=title,
                xlabel=xlabel,
                ylabel=ylabel,
                path=path,
                **kwargs,
            )

    @performance
    class SMAPE:
        def __init__(self, frame, forecast=None, actuals=None, adjust=True, **kwargs):

            self.frame = frame
            self.forecast = forecast
            self.actuals = actuals
            self.adjust = adjust
            self.calculate()

        def calculate(self):

            _set_forecast_and_actuals(self)

            if self.forecast is not None and self.actuals is not None:
                factor = 1 if self.adjust else 2
                self.result = (
                    np.abs(self.actuals - self.forecast)
                    / ((np.abs(self.actuals) + np.abs(self.forecast)) / factor)
                ).sum() / len(self.forecast)
                self.values = np.abs(self.actuals - self.forecast) / (
                    (np.abs(self.actuals) + np.abs(self.forecast)) / factor
                )
            else:
                self.result = None
                self.values = None

        def __repr__(self):
            return f"{self.__class__.__name__}(adjust={self.adjust}, n={self.frame.shape[0]:1.0f}, result={self.result:1.4f})"

        def plot(
            self,
            xlim=(None, None),
            ylim=(None, None),
            bins=20,
            format_xticks="1.0%",
            title=None,
            xlabel=None,
            ylabel=None,
            path=None,
            **kwargs,
        ):
            return _plot(
                self,
                xlim=xlim,
                ylim=ylim,
                bins=bins,
                format_xticks=format_xticks,
                title=title,
                xlabel=xlabel,
                ylabel=ylabel,
                path=path,
                **kwargs,
            )

    class MDA:
        def __init__(self, frame, forecast=None, actuals=None, **kwargs):

            self.frame = frame
            self.forecast = forecast
            self.actuals = actuals
            self.calculate()

        def calculate(self):

            _set_forecast_and_actuals(self)

            if self.forecast is not None and self.actuals is not None:
                self.result = (
                    (
                        (self.forecast < self.forecast.shift(-1)).iloc[:-1]
                        == (self.actuals < self.actuals.shift(-1)).iloc[:-1]
                    )
                    * 1
                ).sum() / (len(self.forecast) - 1)
                self.values = (
                    (self.forecast < self.forecast.shift(-1)).iloc[:-1]
                    == (self.actuals < self.actuals.shift(-1)).iloc[:-1]
                ) * 1
            else:
                self.result = None
                self.values = None

        def __repr__(self):
            return f"{self.__class__.__name__}(n={self.frame.shape[0]:1.0f}, result={self.result:1.4f})"

    @performance
    class MPE:
        def __init__(self, frame, forecast=None, actuals=None, confidence=0.95, **kwargs):

            self.frame = frame
            self.forecast = forecast
            self.actuals = actuals
            self.confidence = confidence
            self.calculate()

        def calculate(self):

            _set_forecast_and_actuals(self)

            if self.forecast is not None and self.actuals is not None:
                self.values = (self.forecast - self.actuals) / self.actuals
                self.result = self.values.mean()
                self.ci_mean = ci_mean(self.values, confidence=self.confidence)
                self.ci_low, self.ci_high = self.ci_mean
                self.bias = not (self.ci_low <= 0 <= self.ci_high)
            else:
                self.result = None
                self.values = None

        def __repr__(self):
            return f"{self.__class__.__name__}(n={self.frame.shape[0]:1.0f}, result={self.result:1.4f}, ci_mean=({self.ci_low:1.4f}, {self.ci_high:1.4f}), bias={self.bias})"

        def plot(
            self,
            xlim=(None, None),
            ylim=(None, None),
            bins=20,
            format_xticks="1.0%",
            title=None,
            xlabel=None,
            ylabel=None,
            path=None,
            **kwargs,
        ):
            return _plot(
                self,
                xlim=xlim,
                ylim=ylim,
                bins=bins,
                format_xticks=format_xticks,
                title=title,
                xlabel=xlabel,
                ylabel=ylabel,
                path=path,
                **kwargs,
            )

    @performance
    class NFM:
        def __init__(self, frame, forecast=None, actuals=None, confidence=0.95, **kwargs):

            self.frame = frame
            self.forecast = forecast
            self.actuals = actuals
            self.confidence = confidence
            self.calculate()

        def calculate(self):

            _set_forecast_and_actuals(self)

            if (self.forecast < 0).any() or (self.actuals < 0).any():
                self.forecast = self.forecast[(self.forecast >= 0) & (self.actuals >= 0)]
                self.actuals = self.actuals[(self.forecast >= 0) & (self.actuals >= 0)]
                warnings.warn(
                    "Forecast and actuals cannot contain negative values. The are deleted."
                )

            if self.forecast is not None and self.actuals is not None:
                self.values = pd.Series(
                    (self.forecast - self.actuals) / (self.forecast + self.actuals),
                    name="NFM values",
                )
                self.result = self.values.mean()
                self.ci_mean = ci_mean(self.values, confidence=self.confidence)
                self.ci_low, self.ci_high = self.ci_mean
                self.bias = not (self.ci_low <= 0 <= self.ci_high)
            else:
                self.result = None
                self.values = None

        def __repr__(self):
            return f"{self.__class__.__name__}(n={self.frame.shape[0]:1.0f}, result={self.result:1.4f}, ci_mean=({self.ci_low:1.4f}, {self.ci_high:1.4f}), bias={self.bias})"

        def plot(
            self,
            xlim=(None, None),
            ylim=(None, None),
            bins=20,
            format_xticks="1.0%",
            title=None,
            xlabel=None,
            ylabel=None,
            path=None,
            **kwargs,
        ):
            return _plot(
                self,
                xlim=xlim,
                ylim=ylim,
                bins=bins,
                format_xticks=format_xticks,
                title=title,
                xlabel=xlabel,
                ylabel=ylabel,
                path=path,
                **kwargs,
            )

def _set_forecast_and_actuals(_obj, **kwargs):
    if isinstance(_obj.frame, pd.Series):
        if isinstance(_obj.forecast, pd.Series):
            _obj.forecast = _obj.forecast
            _obj.actuals = _obj.frame
        elif isinstance(_obj.actuals, pd.Series):
            _obj.forecast = _obj.frame
            _obj.actuals = _obj.actuals
        else:
            raise ValueError(
                "Please provide forecast data or actuals data. One pandas Series is not enough to calculate forecast accuracy."
            )
    elif isinstance(_obj.frame, pd.DataFrame):
        if isinstance(_obj.forecast, str):
            _obj.forecast = _obj.frame[_obj.forecast]
        else:
            _obj.forecast = _obj.frame.iloc[:, 0]

        if isinstance(_obj.actuals, str):
            _obj.actuals = _obj.frame[_obj.actuals]
        else:
            _obj.actuals = _obj.frame.iloc[:, 1]
    return

def _plot(
        plot_obj,
        xlim=(None, None),
        ylim=(None, None),
        bins=20,
        format_stats=".2f",
        format_xticks=None,
        title=None,
        xlabel=None,
        ylabel=None,
        path=None,
        **kwargs,
    ):

        adjusted = " (adjusted)" if hasattr(plot_obj, "adjust") and plot_obj.adjust else ""
        title = title if title is not None else  f"{plot_obj.__class__.__name__}{adjusted}"

        # handle infinity
        values = plot_obj.values.replace([np.inf, -np.inf], np.nan)

        fig, ax = plt.subplots(
            nrows=1, ncols=1, gridspec_kw={"wspace": 0, "hspace": 0}, **kwargs
        )

        # plot
        ax.hist(values, bins=bins, **bluebelt.style("forecast.hist"))

        # ci mean
        if hasattr(plot_obj, "ci_mean"):
            ax.axvspan(
                plot_obj.ci_low,
                plot_obj.ci_high,
                label=f"CI mean ({plot_obj.ci_low:{format_stats}}, {plot_obj.ci_high:{format_stats}})",
                **bluebelt.style("forecast.ci_mean.span"),
            )
            ax.axvline(
                x=plot_obj.ci_low,
                ymin=0,
                ymax=1,
                **bluebelt.style("forecast.ci_mean.line"),
            )
            ax.axvline(
                x=plot_obj.ci_high,
                ymin=0,
                ymax=1,
                **bluebelt.style("forecast.ci_mean.line"),
            )
            ax.legend()

        if hasattr(plot_obj, "bias"):
            if plot_obj.bias:
                t = (
                    "forecast is biased: forecast > actuals"
                    if plot_obj.result > 0
                    else "forecast is biased: forecast < actuals"
                )
            else:
                t = "forecast is not biased"
            ax.text(
                0.02,
                0.98,
                t,
                transform=ax.transAxes,
                **bluebelt.style("forecast.bias.text"),
            )

        # set xlim
        xlim_lower, xlim_upper = xlim
        xlim_lower = xlim_lower or math.floor(values.min())
        xlim_upper = xlim_upper or math.ceil(values.max())
        xlim = (xlim_lower, xlim_upper)
        ax.set_xlim(xlim)

        # remove yticks
        ax.set_yticks([])

        plt.tight_layout()

        # limit axis
        ax.set_ylim(ylim)

        # format ticks
        if format_xticks:
            ax.set_xticks(ax.get_xticks())
            ax.set_xticklabels([f"{x:{format_xticks}}" for x in ax.get_xticks()])

        # labels
        if title:
            ax.set_title(title)
        if xlabel:
            ax.set_xlabel(xlabel)
        if ylabel:
            ax.set_ylabel(ylabel)

        # file
        if path:
            if len(os.path.dirname(path)) > 0 and not os.path.exists(os.path.dirname(path)):
                os.makedirs(os.path.dirname(path))
            plt.savefig(path)
            plt.close()
        else:
            plt.close()
            return fig



import bluebelt
import pandas as pd
import numpy as np
import datetime
import bluebelt.helpers.date
import bluebelt.helpers.language

from collections.abc import Iterable


def _series_function(x):
    function = (
        np.random.uniform(low=0.000014, high=0.000018)
        * (x - np.random.uniform(low=150, high=200)) ** 3
        - np.random.uniform(low=0.0002, high=0.0004)
        * (x - np.random.uniform(low=500, high=750)) ** 2
        - np.random.uniform(low=0.2, high=0.8) * x
        + np.random.uniform(low=500, high=600)
    )
    return function


def series(
    years: int = 1, resolution: float = 0.25, frac: float = 0.02, name: str = "hours"
) -> pd.Series:

    # get the current year
    this_year = datetime.datetime.now().isocalendar()[0]

    # trends
    trends = np.random.normal(loc=1, scale=0.1, size=years)

    data = []

    for year in range(years - 1, -1, -1):
        start_date = datetime.datetime.fromisocalendar(this_year - year, 1, 1)

        nrows = bluebelt.helpers.date.last_iso_week_in_year(this_year - year) * 7
        series = pd.Series(
            data=np.array(list(map(_series_function, np.arange(0, nrows))))
            + (
                np.random.normal(
                    loc=100 * trends[year], scale=4 * trends[year], size=nrows
                )
            ),
            index=pd.date_range(start=start_date, periods=nrows),
        )

        # week & month pattern
        w = pd.Series(
            data=[1.09, 1.05, 0.96, 0.97, 1.11, 0.32, 0.05] * int(nrows / 7),
            index=pd.date_range(start=start_date, periods=nrows),
        )
        m = pd.Series(
            data=series.index.day.map(
                {27: 1.03, 28: 1.06, 29: 1.11, 30: 1.16, 31: 1.09}
            ).fillna(1),
            index=pd.date_range(start=start_date, periods=nrows),
        )

        # set to startpoint
        if year < years - 1:
            series = series + startpoint - series[0]
        startpoint = series[-1]

        # set level
        series = series + np.random.uniform(-100,400)

        series = series.multiply(w).multiply(m)

        # apply resolution
        series = series.divide(resolution).round().multiply(resolution)

        series = series.rename(name)

        data += [series]

    series = pd.concat(data, axis=0)

    # break the series
    series.loc[series.sample(frac=frac).index] = np.nan

    return series


def frame(
    years: int = 1, skills: int = 4, resolution: float = 0.25, frac: float = 0.02
) -> pd.DataFrame:
    data = []
    for skill in range(skills):
        data += [
            series(
                years=years,
                resolution=resolution,
                frac=frac,
                name="skill " + chr(ord("A") + skill),
            )
        ]

    frame = pd.concat(data, axis=1)

    return frame


def holiday_dict(values=None):

    hd = bluebelt.helpers.language.holidays.get(bluebelt.get_language())

    if isinstance(values, Iterable) and (len(values) != len(hd.keys())):
        values = (values * len(hd.keys()))[: len(hd.keys())]

    if isinstance(values, str) or not isinstance(values, Iterable):
        values = [values] * len(hd.keys())

    return {key: values[id] for id, key in enumerate(hd.keys())}


def holiday_list():

    hd = bluebelt.helpers.language.holidays.get(bluebelt.get_language())

    return [val for val in hd.keys()]



import os

import pandas as pd
import numpy as np
import scipy.stats as stats
import matplotlib.pyplot as plt

import bluebelt.helpers.text


class Correlation:
    def __init__(self, frame, columns=None, confidence=0.95, **kwargs):

        # check arguments
        if not isinstance(frame, pd.DataFrame):
            raise ValueError("frame is not a Pandas DataFrame")

        if columns is not None and not isinstance(columns, list):
            raise ValueError("columns is not a list")

        self.frame = frame if columns is None else frame[columns]
        self.columns = columns if columns is not None else self.frame.columns.to_list()
        self.nrows = self.frame.shape[0]
        self.ncols = self.frame.shape[1]
        self.confidence = confidence

        self.calculate()

    def calculate(self):
        self.z = stats.norm.ppf(1 - (1 - self.confidence) / 2)
        self.rci = self._r_ci()

    def __str__(self):
        precision = 3

        result = f"Pearson correlation coefficient with {self.confidence*100:1.1f}% confidence interval\n\n"
        result += (
            self._r_ci()
            .applymap(
                lambda x: np.round(x, precision)
                if isinstance(x, (np.floating, int, float, tuple))
                else x
            )
            .to_string()
        )

        return result

    def __repr__(self):
        return f"{self.__class__.__name__}(columns={self.columns}, nrows={self.nrows}, ncols={self.ncols}, confidence={self.confidence})"

    @property
    def result(self, precision=3):
        print(
            self._r_ci()
            .applymap(
                lambda x: np.round(x, precision)
                if isinstance(x, (np.floating, int, float, tuple))
                else x
            )
            .to_string()
        )

    def plot(self, format_stats="1.2f", title=None, path=None, **kwargs):

        title = (
            title
            or f"correlation plot for {bluebelt.helpers.text._get_nice_list(self.columns)}"
        )

        fig, ax = plt.subplots(nrows=self.ncols, ncols=self.ncols, **kwargs)
        for row, row_name in enumerate(self.columns):
            for col, col_name in enumerate(self.columns):
                if row == col:
                    # plot a histogram where row and column are equal
                    ax[row, col].hist(
                        self.frame[row_name], **bluebelt.style("correlation.hist")
                    )
                else:
                    # plot a scatter plot
                    ax[row, col].plot(
                        self.frame[col_name],
                        self.frame[row_name],
                        **bluebelt.style("correlation.plot"),
                    )

                    # plot regression line
                    dfx = self.frame[
                        self.frame[col_name].notna() & self.frame[row_name].notna()
                    ]
                    a, b = np.polyfit(dfx[col_name], dfx[row_name], 1)
                    ax[row, col].plot(
                        dfx[col_name],
                        (a * dfx[col_name] + b),
                        **bluebelt.style("correlation.line"),
                    )

                    # fetch r and ci
                    r_value = self.rci.iloc[row, col * 2]
                    rho_l = self.rci.iloc[row, 1 + col * 2][0]
                    rho_u = self.rci.iloc[row, 1 + col * 2][1]
                    ax[row, col].text(
                        0.02,
                        0.98,
                        f"r :{r_value:{format_stats}} CI: ({rho_l:{format_stats}} - {rho_u:{format_stats}})",
                        transform=ax[row, col].transAxes,
                        **bluebelt.style("correlation.text"),
                    )

                    # share X-axis
                    if col == 0 and row > 1:
                        ax[row, col].sharex(ax[1, col])
                    elif col > 0 and row > 0:
                        ax[row, col].sharex(ax[0, col])

                    # share Y-axis
                    if row == 0 and col > 1:
                        ax[row, col].sharey(ax[row, 1])
                    elif row > 0 and col > 0:
                        ax[row, col].sharey(ax[row, 0])

                if row == self.ncols - 1:
                    ax[row, col].set_xlabel(col_name)

                if col == 0:
                    ax[row, col].set_ylabel(row_name)

                ax[row, col].set_xticks([])
                ax[row, col].set_yticks([])

        # labels
        if title:
            ax[0, 0].set_title(title)

        plt.tight_layout()
        plt.subplots_adjust(hspace=0)
        plt.subplots_adjust(wspace=0)

        # file
        if path:
            if len(os.path.dirname(path)) > 0 and not os.path.exists(
                os.path.dirname(path)
            ):
                os.makedirs(os.path.dirname(path))
            plt.savefig(path)
            plt.close()
        else:
            plt.close()
            return fig

    def _r_ci(self):

        cols = []
        for col in range(0, self.ncols):
            rows = []
            for row in range(0, self.ncols):

                # calculate r and ci
                if row != col:
                    frame = self.frame[[self.columns[row], self.columns[col]]].dropna(
                        axis=0
                    )

                    r_value = np.corrcoef(
                        frame[self.columns[row]], frame[self.columns[col]]
                    )[0, 1]
                    z_l = 0.5 * (np.log(1 + r_value) - np.log(1 - r_value)) - self.z / (
                        frame.shape[0] - 3
                    )
                    z_u = 0.5 * (np.log(1 + r_value) - np.log(1 - r_value)) + self.z / (
                        frame.shape[0] - 3
                    )
                    rho_l = (np.exp(2 * z_l) - 1) / (np.exp(2 * z_l) + 1)
                    rho_u = (np.exp(2 * z_u) - 1) / (np.exp(2 * z_u) + 1)

                    ci = [r_value, (rho_l, rho_u)]
                else:
                    ci = ["", ""]  # [1., 1., 1.]

                rows += ci
            cols.append(rows)

        columns_array = np.array(
            [
                item
                for sublist in [[col] * 2 for col in self.frame.columns]
                for item in sublist
            ]
            + ["r", "CI"] * self.ncols
        ).reshape(2, self.ncols * 2)
        columns = pd.MultiIndex.from_tuples(
            list(zip(*columns_array))
        )  # , names=["array", "value"])

        return pd.DataFrame(cols, index=self.columns, columns=columns)

    def _ci(self):

        cols = []
        for col in range(0, self.ncols):
            rows = []
            for row in range(0, self.ncols):

                # calculate r and ci
                if row != col:
                    frame = self.frame[[self.columns[row], self.columns[col]]].dropna(
                        axis=0
                    )

                    r_value = np.corrcoef(
                        frame[self.columns[row]], frame[self.columns[col]]
                    )[0, 1]
                    z_l = 0.5 * (np.log(1 + r_value) - np.log(1 - r_value)) - self.z / (
                        frame.shape[0] - 3
                    )
                    z_u = 0.5 * (np.log(1 + r_value) - np.log(1 - r_value)) + self.z / (
                        frame.shape[0] - 3
                    )
                    rho_l = (np.exp(2 * z_l) - 1) / (np.exp(2 * z_l) + 1)
                    rho_u = (np.exp(2 * z_u) - 1) / (np.exp(2 * z_u) + 1)

                    ci = (rho_l, rho_u)
                else:
                    ci = (1.0, 1.0)

                rows.append(ci)
            cols.append(rows)

        return pd.DataFrame(cols, index=self.columns, columns=self.columns)

    def _r(self):
        return pd.DataFrame(
            [
                [
                    (np.corrcoef(self.frame[col], self.frame[row])[0, 1])
                    for col in self.columns
                ]
                for row in self.columns
            ],
            index=self.columns,
            columns=self.columns,
        )



weekdays = {
    "nl": {0:"ma", 1: "di", 2: "wo", 3: "do", 4: "vr", 5: "za", 6: "zo"},
    "en": {0:"Mo", 1: "Tu", 2: "We", 3: "Th", 4: "Fr", 5: "Sa", 6: "Su"},
}

holidays = {
    "nl": {
        "nieuwjaarsdag": "_new_years_day",
        "goede vrijdag": "_good_friday",
        "eerste paasdag": "_easter",
        "tweede paasdag": "_easter_monday",
        "hemelvaartsdag": "_ascension_day",
        "eerste pinksterdag": "_pentecost",
        "tweede pinksterdag": "_whit_monday",
        "eerste kerstdag": "_christmas",
        "tweede kerstdag": "_boxing_day",
        "koningsdag": "_kings_day",
        "bevrijdingsdag": "_liberation_day",
    },
    "en": {
        "new years day": "_new_years_day",
        "good friday": "_good_friday",
        "easter": "_easter",
        "easter monday": "_easter_monday",
        "ascension day": "_ascension_day",
        "pentecost": "_pentecost",
        "whit monday": "_whit_monday",
        "christmas": "_christmas",
        "boxing day": "_boxing_day",
        "kings day": "_kings_day",
        "liberationday": "_liberation_day",
    }
}